#include "jex.h"
#include <Xm/List.h>
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>

Widget bd=NULL,bdl;
extern Widget top;
extern Pixmap pixmap;

void OpenBookmark(Widget w,Editor *e,void *cbs)
{
  XmString xstr;
  char *str;
  
SP("OpenBookmark");
  XtVaGetValues(w,XmNlabelString,&xstr,NULL);
  XmStringGetLtoR(xstr,XmSTRING_DEFAULT_CHARSET,&str);
  XmStringFree(xstr);
  if (e->new_window)
  {
    CreateNewEditor();
    if (loadfile(editor[editors-1],str,0,1))
      LoadError(editor[editors-1]);
  }
  else
  {
    db_strcpy(e->choice,str);
    if (e->saved)
      LoadOldFile(e);
    else
      AskToSave(e,(void*)LoadOldFile);
  }
EP;
}

void UpdateEditorBookmarks(Editor *e)
{
  int count,x;
  XmStringTable items;

  SP("UpdateEditorBookmarks");
  XtVaGetValues(bdl,XmNitemCount,&count,XmNitems,&items,NULL);
  for (x=0;x<count && x<MAXBOOKMARKBUTTONS;x++)
  {
    e->bookm[x]=XtVaCreateManagedWidget("bookmarkbutton",
      xmPushButtonGadgetClass,e->bookmarkmenu,XmNlabelString,items[x],NULL);
    XtAddCallback(e->bookm[x],XmNactivateCallback,(XtCallbackProc)OpenBookmark,(void*)e);
  }
  EP;
}

void DeleteEditorBookmarks(Editor *e)
{
  int count,x;

  SP("DeleteEditorBookmarks");
  XtVaGetValues(bdl,XmNitemCount,&count,NULL);
  for (x=0;x<count && x<MAXBOOKMARKBUTTONS;x++)
    XtDestroyWidget(e->bookm[x]);
  EP;
}

void CloseBookmark(Widget w,void *e,void *cbs)
{
SP("CloseBookmark");
  XtPopdown(bd);
EP;
}

void GotoBookmark(Widget w,void *date,XmListCallbackStruct *cbs)
{
  XmStringTable selected_items;
  int num_selected;
  char *str;
  int x;
  
SP("GotoBookmark");
  XtVaGetValues(bdl,XmNselectedItemCount,&num_selected,XmNselectedItems,&selected_items,NULL);
  if (num_selected==0)
  {
    EP;
    return;
  }
  XmStringGetLtoR(selected_items[0],XmSTRING_DEFAULT_CHARSET,&str);
  for (x=0;x<editors;x++)
  {
    if (db_strcmp(editor[x]->filename,str)==0)
    {
      XMapRaised(XtDisplay(editor[x]->shell),XtWindow(editor[x]->shell));
      XtFree(str);
      EP;
      return;
    }
  }
  CreateNewEditor();
  if (loadfile(editor[editors-1],str,0,1))
    LoadError(editor[editors-1]);
  XtFree(str);
EP;
}

void DeleteBookmark(Widget w,void *data,void *cbs)
{
  int *pos,count,x;
  
SP("DeleteBookmark");
  XmListGetSelectedPos(bdl,&pos,&count);
  if (count)
  {
    for (x=0;x<editors;x++)
      DeleteEditorBookmarks(editor[x]);
    XmListDeletePos(bdl,pos[0]);
    XtFree((char*)pos);
    for (x=0;x<editors;x++)
      UpdateEditorBookmarks(editor[x]);
  }
EP;
}
   
void CreateBookmarkDialog(void)
{
  Widget w;
  Arg arg[2];

SP("CreateBookmarkDialog");
  if (bd==NULL)
  {
    bd=XtVaAppCreateShell("BookmarkDialog","BookmarkDialog",topLevelShellWidgetClass,XtDisplay(top),
  	XmNallowShellResize,True,
    	XmNiconPixmap,pixmap,
  	NULL);
    w=XmCreateMessageBox(bd,"BookmarkDialog",NULL,0);
    XtSetArg(arg[0],XmNlistSizePolicy,XmCONSTANT);
    bdl=XmCreateScrolledList(w,"BookmarkList",arg,1);
    XtVaSetValues(XtParent(bdl),
    	XmNscrollBarDisplayPolicy,XmSTATIC, 
    	NULL);
    XtVaSetValues(bdl,
    	XmNscrollBarDisplayPolicy,XmSTATIC, 
    	XmNlistSizePolicy,XmCONSTANT, 
    	NULL);
    XtAddCallback(bdl,XmNdefaultActionCallback,(XtCallbackProc)GotoBookmark,NULL);
    XtAddCallback(w,XmNokCallback,(XtCallbackProc)CloseBookmark,NULL);
    XtAddCallback(w,XmNcancelCallback,(XtCallbackProc)GotoBookmark,NULL);
    XtAddCallback(w,XmNhelpCallback,(XtCallbackProc)DeleteBookmark,NULL);
    XtManageChild(bdl);
    XtManageChild(w);
  }
EP;
}

void AddBookmark(char *str)
{
  XmString xstr;
  int x;
  
SP("AddBookmark");
  if (bdl && str[0])
  {
    for (x=0;x<editors;x++)
      DeleteEditorBookmarks(editor[x]);
    xstr=XmStringCreateLocalized(str);
    if (!XmListItemExists(bdl,xstr))
      XmListAddItemUnselected(bdl,xstr,0);
    XmStringFree(xstr);
    for (x=0;x<editors;x++)
      UpdateEditorBookmarks(editor[x]);
  }
EP;
}

void ShowBookmarkDialog(void)
{
SP("ShowBookmarkDialog");
  XtPopup(bd,XtGrabNone);
  XMapRaised(XtDisplay(bd),XtWindow(bd));
EP;
}

