#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <stdarg.h>
#include <fcntl.h>
#include "dbgfunc.h"
#include "hashtab.h"
#include "dbgstack.h"

#define PE(str) {if (file[0]) {printf("file: %s, ",file);} if (function[0]) {printf("function: %s, ",function);} if (line) {printf("line: %d ",line);} printf("-> %s\n",str);_db_log("%s",str);}
#define BITSET(n,x) (x[n/(sizeof(int)*8)]&(1<<(n%(sizeof(int)*8))))

#define _FILE 1
#define _SOCKET 2
#define _PIPE 4
#define _DIR 8

typedef struct {
		 void *ptr;
		 int size;
		 int line;
		 char file[20];
	       } Pointer;

typedef struct {
		 FILE *f;
		 char type;
		 int line;
		 char file[20];
	       } FilePointer;

typedef struct {
		 int f;
		 char type;
		 int line;
		 char file[20];
	       } FileDescriptor;

HashTable ht,ht2,ht3;
char initialized=0,db_debug_mode=1;
char db_verbose=0,db_postmortemdump=1;
unsigned _db_totsize=0;
extern int errno;
char _db_prgname[255]="";
void (*_db_sighand)(int)=NULL;
char _db_logfile=0;
char _db_logfilename[255]="";
char _db_lastfile[20]="";
int _db_lastline=0;
int _db_limitlog=0;
char _db_lastfunction[32]="";
Stack *_db_stack=NULL;

#define _db_add_filepointer(point,t) {FilePointer p;p.f=point;p.type=t;p.line=line;strncpy(p.file,file,19);HTAddRecord(ht2,&p);}
#define _db_add_filedescriptor(point,t) {FileDescriptor p;p.type=t;p.f=point;p.line=line;strncpy(p.file,file,19);HTAddRecord(ht3,&p);}
#define _db_initialize {strncpy(_db_lastfile,file,19); \
			_db_lastline=line; \
			if (!initialized) { \
			 ht=HTCreate(1024,sizeof(void*),sizeof(Pointer)); \
			 ht2=HTCreate(32,sizeof(FILE *),sizeof(FilePointer)); \
			 _db_add_filepointer(stdout,_FILE) \
			 _db_add_filepointer(stderr,_FILE) \
			 _db_add_filepointer(stdin,_FILE) \
			 ht3=HTCreate(32,sizeof(int),sizeof(FileDescriptor)); \
			 _db_add_filedescriptor(0,_FILE) \
			 _db_add_filedescriptor(1,_FILE) \
			 _db_add_filedescriptor(2,_FILE) \
			 _db_stack=_db_StackCreate(); \
			 initialized=1; \
			}}
#define _db_add_pointer(point,sz) {Pointer p;p.ptr=point;p.size=sz;p.line=line;strncpy(p.file,file,19);HTAddRecord(ht,&p);}
#define _db_check_pointer(ptr) HTLookup(ht,&ptr)
#define _db_delete_pointer(ptr) HTDeleteRecord(ht,&ptr)
#define _db_checksize(ptr,sz,f,a) {Pointer p; \
			           if (HTGetRecord(ht,&ptr,&p)) { \
			             if (p.size<sz) \
			             {if (file[0]) {printf("file: %s, ",file);} if (function[0]) {printf("function: %s, ",function);} if (line) {printf("line: %d ",line);} printf("-> %s, argument %d: size of pointer too small(%d)\n",f,a,p.size);_db_log("%s, argument %d: size of pointer too small(%d)",f,a,p.size);} \
			           }}
#define _db_perror(str) {if (db_debug_mode) {if (file[0]) {printf("file: %s, ",file);} if (function[0]) {printf("function: %s, ",function);} if (line) {printf("line: %d ",line);} printf("-> %s: %s\n",str,strerror(errno));_db_log("%s: %s",str,strerror(errno));}}

int _db_check_filepointer(FILE *f,char t)
{
  FilePointer p;
  
  if (!HTGetRecord(ht2,&f,&p))
    return 0;
  return (p.type&t);
}

int _db_check_filedescriptor(int f,char t)
{
  FileDescriptor p;
  
  if (!HTGetRecord(ht3,&f,&p))
    return 0;
  return (p.type&t);
}

int _db_delete_filepointer(FILE *f,char t)
{
  FilePointer p;
  
  if (!HTGetRecord(ht2,&f,&p))
    return 0;
  if  (!(p.type&t))
    return 0;
  return HTDeleteRecord(ht2,&f);
}

int _db_delete_filedescriptor(int f,char t)
{
  FileDescriptor p;
  
  if (!HTGetRecord(ht3,&f,&p))
    return 0;
  if  (!(p.type&t))
    return 0;
  return HTDeleteRecord(ht3,&f);
}

void PrintMem(Pointer *p)
{
  printf("  Memory allocated in file %s, line %d, size: %d\n",p->file,p->line,p->size);
  if (_db_logfile)
    _db_log2("  Memory allocated in file %s, line %d, size: %d\n",p->file,p->line,p->size);
  _db_totsize+=p->size;
}

void PrintFiles(FilePointer *p)
{
  if (p->f!=stdout && p->f!=stderr && p->f!=stdin)
  {
    if (p->type&_FILE)
    {
      printf("  File pointer opened in file %s, line %d\n",p->file,p->line);
      if (_db_logfile)
        _db_log2("  File pointer opened in file %s, line %d\n",p->file,p->line);
    }
    else if (p->type&_PIPE)
    {
      printf("  Pipe opened in file %s, line %d\n",p->file,p->line);
      if (_db_logfile)
        _db_log2("  Pipe opened in file %s, line %d\n",p->file,p->line);
    }
    else if (p->type&_DIR)
    {
      printf("  Directory opened in file %s, line %d\n",p->file,p->line);
      if (_db_logfile)
        _db_log2("  Directory opened in file %s, line %d\n",p->file,p->line);
    }
  }
}

void PrintFD(FileDescriptor *p)
{
  if (p->f>2)
  {
    if (p->type&_FILE)
    {
      printf("  File descriptor (%d) opened in file %s, line %d\n",p->f,p->file,p->line);
      if (_db_logfile)
        _db_log2("  File descriptor (%d) opened in file %s, line %d\n",p->f,p->file,p->line);
    }
    else
    {
      printf("  Socket (%d) opened in file %s, line %d\n",p->f,p->file,p->line);
      if (_db_logfile)
        _db_log2("  Socket (%d) opened in file %s, line %d\n",p->f,p->file,p->line);
    }
  }
}

void FinalCheck(void)
{
  int x=0;
  
  if (!db_postmortemdump)
    return;
  if (HTNumberOfRecords(ht))
  {
    printf("\n------- Post-execution diagnostics --------\n");
    if (_db_logfile)
      _db_log2("\n------- Post-execution diagnostics --------\n");
    x=1;
    printf("\n* Memory leaks:\n\n");
    if (_db_logfile)
      _db_log2("\n* Memory leaks:\n\n");
    HTForEach(ht,(HTForEachProc)PrintMem);
    printf("  ------------------> total: %u bytes (%.2f K)\n",_db_totsize,(double)_db_totsize/1024.0);
    if (_db_logfile)
      _db_log2("  ------------------> total: %u bytes (%.2f K)\n",_db_totsize,(double)_db_totsize/1024.0);
  }
  if (HTNumberOfRecords(ht2)>3)
  {
    if (!x)
    {
      printf("\n------- Post-execution diagnostics -------\n");
      if (_db_logfile)
        _db_log2("\n------- Post-execution diagnostics -------\n");
      x=1;
    }   
    printf("\n* Open file pointers, directories or pipes:\n\n");
    if (_db_logfile)
      _db_log2("\n* Open file pointers, directories or pipes:\n\n");
    HTForEach(ht2,(HTForEachProc)PrintFiles);
  }
  if (HTNumberOfRecords(ht3)>3)
  {
    if (!x)
    {
      printf("\n------- Post-execution diagnostics -------\n");
      if (_db_logfile)
        _db_log2("\n------- Post-execution diagnostics -------\n");
      x=1;
    }   
    printf("\n* Open file descriptors or sockets:\n\n");
    if (_db_logfile)
      _db_log2("\n* Open file descriptors or sockets:\n\n");
    HTForEach(ht3,(HTForEachProc)PrintFD);
  }
}

void LimitLog(void)
{
  FILE *f,*f2;
  int n,l=0,x;
  char buf[5000];
  
  if (_db_logfile && _db_limitlog)
  {
    if ((f=fopen(_db_logfilename,"r"))==NULL)
      return;
    while (fgets(buf,5000,f))
      l++;
    if (l<=_db_limitlog-1)
    {
      fclose(f);
      return;
    }
    fseek(f,0,0);
    if ((f2=tmpfile())==NULL)
    {
      fclose(f);
      return;
    }
    x=0;
    while (fgets(buf,5000,f))
    {
/*      buf[strlen(buf)-1]=0;*/
      if (x++>l-_db_limitlog+1)
        fputs(buf,f2);
    }
    fclose(f);
    if ((f=fopen(_db_logfilename,"w+"))==NULL)
    {
      fclose(f2);
      return;
    }
    fseek(f2,0,0);
    while (fgets(buf,5000,f2))
    {
/*      buf[strlen(buf)-1]=0;*/
      fputs(buf,f);
    }
    fclose(f);
    fclose(f2);
  }
}

void _db_log2(char *str,...)
{
  va_list app;
  char s[1000];
  FILE *f;

  if (_db_logfile)
  {
    LimitLog();
    if ((f=fopen(_db_logfilename,"a+"))==NULL)
      return;
    va_start(app,str);
    vsprintf(s,str,app);
    va_end(app);
    fprintf(f,"%s",s);
    fclose(f);
  }
}

void _db_log(char *str,...)
{
  va_list app;
  time_t t;
  int line=_db_lastline;
  char timebuf[100],s[1000];
  char *file=_db_lastfile,*function=_db_lastfunction;
  FILE *f;

  if (str==NULL)
    PE("argument 1 to Log is NULL")
  if (_db_logfile)
  {
    LimitLog();
    if ((f=fopen(_db_logfilename,"a+"))==NULL)
      return;
    va_start(app,str);
    vsprintf(s,str,app);
    va_end(app);
    time(&t);
    strftime(timebuf,100,"%d-%m-%Y %H:%M:%S",localtime(&t));
    if (function[0])
      fprintf(f,"%19s (%s:%s():%d) -> %s\n",timebuf,file,function,line,s);
    else
      fprintf(f,"%19s (%s:%d) -> %s\n",timebuf,file,line,s);
    fclose(f);
  }
}

char *_db_strsignal(int sig)
{
  switch (sig)
  {
    case SIGQUIT: return "Quit";
    case SIGILL:  return "Illegal instruction";
    case SIGTRAP: return "Trap";
    case SIGIOT:  return "I/O Trap";
    case SIGFPE:  return "Floating point exception";
    case SIGKILL: return "Kill";
    case SIGBUS:  return "Bus error";
    case SIGSEGV: return "Segmentation violation";
    case SIGTERM: return "Terminate";
    case SIGHUP:  return "Hangup";
    case SIGINT:  return "Interrupt";
    case SIGIO:   return "I/O error";
    case SIGPIPE: return "Pipe error";
    case SIGUSR1: return "User 1 signal";
    case SIGUSR2: return "User 2 signal";
    case SIGPROF: return "Profiling signal";
  }
  return "Unknown signal";
}

void _db_terminate(int sig)
{
  signal(sig,SIG_IGN);
  printf("\n%s: Received signal %d -> %s\n",_db_prgname,sig,_db_strsignal(sig));
  _db_log("\n%s: Received on signal %d -> %s\n",_db_prgname,sig,_db_strsignal(sig));
  if (_db_lastfile[0] && _db_lastline)
  {
    printf("\nLast known location -> file %s, line %d\n",_db_lastfile,_db_lastline);
    _db_log("\nLast known location -> file %s, line %d\n",_db_lastfile,_db_lastline);
  }
  _db_StackDisplay(_db_stack);
  if (_db_sighand!=NULL)
    _db_sighand(sig);
  exit(sig);
}

void _db_atexit(void)
{
  FinalCheck();
}

void DBInit(char *program_name,char *logfile,int limitlog,void (*SigHand)(int))
{
  char file[5]="";
  int line=0;
  
  _db_initialize;
  if (program_name==NULL)
    strcpy(_db_prgname,"");
  else
    strncpy(_db_prgname,program_name,255);
  _db_sighand=SigHand;
  if (logfile!=NULL && logfile[0])
  {
    _db_logfile=1;
    strncpy(_db_logfilename,logfile,255);
    _db_limitlog=(limitlog>=0)?(limitlog):0;
  }
  signal(SIGQUIT,_db_terminate);
  signal(SIGILL,_db_terminate);
  signal(SIGTRAP,_db_terminate);
  signal(SIGIOT,_db_terminate);
  signal(SIGFPE,_db_terminate);
  signal(SIGKILL,_db_terminate);
  signal(SIGBUS,_db_terminate);
  signal(SIGSEGV,_db_terminate);
  signal(SIGTERM,_db_terminate);
  signal(SIGHUP,_db_terminate);
  signal(SIGINT,_db_terminate);
  signal(SIGIO,_db_terminate);
  signal(SIGPIPE,_db_terminate);
  signal(SIGUSR1,SIG_IGN);
  signal(SIGUSR2,SIG_IGN);
  signal(SIGPROF,SIG_IGN);
  signal(SIGABRT,_db_terminate);
  atexit(_db_atexit);
}


void *_db_malloc(int size,char *file,char *function,int line)
{
  void *tmp;
  
  _db_initialize;
  if (!size)
    PE("argument to malloc is 0");
  if ((tmp=malloc(size))==NULL)
  {
    _db_perror("malloc")
  }
  else
    _db_add_pointer(tmp,size);
  return tmp;
}
  
void _db_free(void *ptr,char *file,char *function,int line)
{
  _db_initialize;
  if (!ptr)
    PE("argument to free is NULL (skipped)")
  else
    if (!_db_delete_pointer(ptr))
      PE("free of an unknown pointer")
  free(ptr);
}
  
void *_db_realloc(void *ptr,int size,char *file,char *function,int line)
{
  _db_initialize;
  if (!ptr && size==0)
    PE("arguments to realloc are NULL,0 (skipped)")
  else
  {
    if (ptr)
    {
      if (!_db_delete_pointer(ptr))
        PE("realloc of an unknown pointer")
    }
    ptr=realloc(ptr,size);
    if (!ptr && size)
      _db_perror("realloc")
    else if (ptr && !size)
      PE("realloc returned not NULL pointer while size=0")
    if (ptr)
      _db_add_pointer(ptr,size);
  }  
  return ptr;
}
  

void *_db_calloc(int size,int n,char *file,char *function,int line)
{
  void *tmp;
  
  _db_initialize;
  if (!size)
    PE("argument 1 to calloc is 0")
  if (!n)
    PE("argument 2 to calloc is 0")
  if ((tmp=calloc(size,n))==NULL)
  {
    _db_perror("calloc")
  }
  else
    _db_add_pointer(tmp,size*n);
  return tmp;
}
  
char *_db_strcpy(char *str1,char *str2,char *file,char *function,int line)
{
  char *tmp;

  _db_initialize;
  if (str1==NULL)
    PE("argument 1 to strcpy is NULL")
  if (str2==NULL)
    PE("argument 2 to strcpy is NULL")
  _db_checksize(str1,strlen(str2)+1,"strcpy",1)
  if ((tmp=strcpy(str1,str2))==NULL)
    _db_perror("strcpy")
  return tmp;
}
  
char *_db_strncpy(char *str1,char *str2,int n,char *file,char *function,int line)
{
  char *tmp;

  _db_initialize;
  if (str1==NULL)
    PE("argument 1 to strncpy is NULL")
  if (str2==NULL)
    PE("argument 2 to strncpy is NULL")
  if (!n)
    PE("argument 3 to strncpy is 0")
  _db_checksize(str1,n,"strncpy",1)
  if ((tmp=strncpy(str1,str2,n))==NULL)
    _db_perror("strncpy")
  return tmp;
}  

int _db_strcmp(char *str1,char *str2,char *file,char *function,int line)
{
  _db_initialize;
  if (str1==NULL)
    PE("argument 1 to strcmp is NULL")
  if (str2==NULL)
    PE("argument 2 to strcmp is NULL")
  return strcmp(str1,str2);
}
  

char *_db_strdup(char *str1,char *file,char *function,int line)
{
  char *tmp;

  _db_initialize;
  if (str1==NULL)
    PE("argument 1 to strdup is NULL")
  if ((tmp=strdup(str1))==NULL)
    _db_perror("strdup")
  else
    _db_add_pointer(tmp,sizeof(str1+1));
  return tmp;
}

void *_db_memchr(void *ptr,int chr,unsigned int n,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to memchr is NULL")
  _db_checksize(ptr,n,"memchr",1)
  return memchr(ptr,chr,n);
}
    
int _db_memcmp(void *ptr1,void *ptr2,unsigned int n,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr1==NULL)
    PE("argument 1 to memcmp is NULL")
  if (ptr2==NULL)
    PE("argument 2 to memcmp is NULL")
  if (!n)
    PE("argument 3 to memcmp is NULL")
  _db_checksize(ptr1,n,"memcmp",1)
  _db_checksize(ptr2,n,"memcmp",2)
  return memcmp(ptr1,ptr2,n);
}
    
void *_db_memcpy(void *ptr1,void *ptr2,unsigned int n,char *file,char *function,int line)
{
  void *tmp;
  
  _db_initialize;
  if (ptr1==NULL)
    PE("argument 1 to memcpy is NULL")
  if (ptr2==NULL)
    PE("argument 2 to memcpy is NULL")
  if (!n)
    PE("argument 3 to memcpy is NULL")
  _db_checksize(ptr1,n,"memcpy",1)
  _db_checksize(ptr2,n,"memcpy",2)
  if ((tmp=memcpy(ptr1,ptr2,n))==NULL)
    _db_perror("memcpy")
  return tmp;
}

void *_db_memmove(void *ptr1,void *ptr2,unsigned int n,char *file,char *function,int line)
{
  void *tmp;
  
  _db_initialize;
  if (ptr1==NULL)
    PE("argument 1 to memmove is NULL")
  if (ptr2==NULL)
    PE("argument 2 to memmove is NULL")
  if (!n)
    PE("argument 3 to memmove is NULL")
  _db_checksize(ptr1,n,"memmove",1)
  _db_checksize(ptr2,n,"memmove",2)
  if ((tmp=memmove(ptr1,ptr2,n))==NULL)
    _db_perror("memmove")
  return tmp;
}

void *_db_memset(void *ptr,int chr,unsigned int n,char *file,char *function,int line)
{
  void *tmp;
  
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to memset is NULL")
  _db_checksize(ptr,n,"memset",1)
  if ((tmp=memset(ptr,chr,n))==NULL)
    _db_perror("memset")
  return tmp;
}

char *_db_strcat(char *ptr1,char *ptr2,char *file,char *function,int line)
{
  void *tmp;
  
  _db_initialize;
  if (ptr1==NULL)
    PE("argument 1 to strcat is NULL")
  if (ptr2==NULL)
    PE("argument 2 to strcat is NULL")
  _db_checksize(ptr1,strlen(ptr1)+strlen(ptr2)+1,"strcat",1)
  if ((tmp=strcat(ptr1,ptr2))==NULL)
    _db_perror("strcat")
  return tmp;
}

char *_db_strchr(char *ptr,int chr,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to strchr is NULL")
  return strchr(ptr,chr);
}

int _db_strcspn(char *ptr1,char *ptr2,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr1==NULL)
    PE("argument 1 to strcspn is NULL")
  if (ptr2==NULL)
    PE("argument 2 to strcspn is NULL")
  return strcspn(ptr1,ptr2);
}

unsigned int _db_strlen(char *ptr,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to strlen is NULL")
  return strlen(ptr);
}

char *_db_strncat(char *ptr,char *ptr2,unsigned int n,char *file,char *function,int line)
{
  void *tmp;
  
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to strncat is NULL")
  if (ptr2==NULL)
    PE("argument 2 to strncat is NULL")
  if (!n)
    PE("argument 3 to strncat is 0")
  _db_checksize(ptr,strlen(ptr)+n+1,"strncat",1)
  if ((tmp=strncat(ptr,ptr2,n))==NULL)
    _db_perror("strncat")
  return tmp;
}

int _db_strncmp(char *ptr,char *ptr2,unsigned int n,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to strncmp is NULL")
  if (ptr2==NULL)
    PE("argument 2 to strncmp is NULL")
  if (!n)
    PE("argument 3 to strncmp in 0")
  return strncmp(ptr,ptr2,n);
}

char *_db_strpbrk(char *ptr,char *ptr2,char *file,char *function,int line)
{
  void *tmp;
  
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to strpbrk is NULL")
  if (ptr2==NULL)
    PE("argument 2 to strpbrk is NULL")
  if ((tmp=strpbrk(ptr,ptr2))==NULL)
    _db_perror("strpbrk")
  return tmp;
}

char *_db_strrchr(char *ptr,int chr,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to strrchr is NULL")
  return strrchr(ptr,chr);
}

int _db_strspn(char *ptr,char *ptr2,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to strspn is NULL")
  if (ptr2==NULL)
    PE("argument 2 to strspn is NULL")
  return strspn(ptr,ptr2);
}

char *_db_strstr(char *ptr,char *ptr2,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to strstr is NULL")
  if (ptr2==NULL)
    PE("argument 2 to strstr is NULL")
  return strstr(ptr,ptr2);
}

char *_db_strtok(char *ptr,char *ptr2,char *file,char *function,int line)
{
  _db_initialize;
  if (ptr==NULL)
    PE("argument 1 to strtok is NULL")
  if (ptr2==NULL)
    PE("argument 2 to strtok is NULL")
  return strtok(ptr,ptr2);
}

void _db_clearerr(FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("clearerr on unknown file")
  clearerr(f);
}
  
int _db_close(int f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_delete_filedescriptor(f,_FILE|_SOCKET))
    PE("close of unknown file")
  if ((tmp=close(f))==-1)
    _db_perror("close")
  return tmp;    
}

int _db_creat(char *f,int mode,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to creat is NULL");
  if ((tmp=creat(f,mode))>=0)
    _db_add_filedescriptor(tmp,_FILE)
  else
    _db_perror("creat")
  return tmp;
}
  
int _db_fclose(FILE *f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_delete_filepointer(f,_FILE))
    PE("fclose of unknown file")
  if ((tmp=fclose(f))==-1)
    _db_perror("fclose")
  return tmp;
}

int _db_feof(FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("feof on unknown file")
  return feof(f);
}

int _db_ferror(FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("ferror on unknown file")
  return ferror(f);
}

int _db_fflush(FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("fflush on unknown file")
  return fflush(f);
}

int _db_fgetc(FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("fgetc on unknown file")
  return fgetc(f);
}

char *_db_fgets(char *str,int num,FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("fgets on unknown file")
  if (str==NULL)
    PE("argument 1 to fgets is null")
  if (num<0)
    PE("argument 2 to fgets < 0")
  return fgets(str,num,f);
}

FILE *_db_fopen(char *f,char *mode,char *file,char *function,int line)
{
  FILE *tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to fopen is NULL")
  if (mode==NULL)
    PE("argument 2 to fopen is NULL")
  if ((tmp=fopen(f,mode))!=NULL)
    _db_add_filepointer(tmp,_FILE)
  else
    _db_perror("fopen")
  return tmp;
}

int _db_fputc(int chr,FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("fputc on unknown file")
  return fputc(chr,f);
}

int _db_fputs(char *str,FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("fputs on unknown file")
  if (str==NULL)
    PE("argument 1 to fputs is NULL")
  return fputs(str,f);
}

int _db_fread(void *b,int s,int n,FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("fread on unknown file");
  if (b==NULL)
    PE("argument 1 to fread is NULL")
  if (s<0)
    PE("argument 2 to fread < 0")
  if (n<0)
    PE("argument 3 to fread < 0")
  _db_checksize(b,s*n,"fread",1);
  return fread(b,s,n,f);
}

FILE *_db_freopen(char *n,char *m,FILE *f,char *file,char *function,int line)
{
  FILE *tmp;
  
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE))
    PE("freopen on unknown file");
  if (n==NULL)
    PE("argument 1 to freopen is NULL")
  if (m==NULL)
    PE("argument 2 to freopen is NULL")
  if ((tmp=freopen(n,m,f))==NULL)
    _db_perror("freopen")
  else
  {
    if (!_db_check_filepointer(tmp,_FILE))
    {
      _db_delete_filepointer(f,_FILE);
      _db_add_filepointer(tmp,_FILE)
    }
  }  
  return tmp;    
}

int _db_fseek(FILE *f,long off,int or,char *file,char *function,int line)
{
  int tmp;

  _db_initialize;
  if (!_db_check_filepointer(f,_FILE))
    PE("fseek on unknown file");
  if (or!=SEEK_SET && or!=SEEK_CUR && or!=SEEK_END)
    PE("argument 3 to fseek has invalid value")
  if ((tmp=fseek(f,off,or))==-1)
    _db_perror("fseek")
  return tmp;  
}

long _db_ftell(FILE *f,char *file,char *function,int line)
{
  int tmp;

  _db_initialize;
  if (!_db_check_filepointer(f,_FILE))
    PE("ftell on unknown file");
  if ((tmp=ftell(f))==-1)
    _db_perror("ftell")
  return tmp;
}

int _db_fwrite(void *b,int s,int n,FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("fwrite on unknown file");
  if (b==NULL)
    PE("argument 1 to fwrite is NULL")
  if (s<0)
    PE("argument 2 to fwrite < 0")
  if (n<0)
    PE("argument 3 to fwrite < 0")
  _db_checksize(b,s*n,"fwrite",1);
  return fwrite(b,s,n,f);
}

int _db_getc(FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("getc on unknown file")
  return getc(f);
}

long _db_lseek(int f,long off,int or,char *file,char *function,int line)
{
  int tmp;
  
  if (!_db_check_filedescriptor(f,_FILE|_SOCKET))
    PE("seek on unknown file")
  if (or<0 || or >2)
    PE("argument 3 to seek has invalid value")
  if ((tmp=lseek(f,off,or))==-1)
    _db_perror("lseek")
  return tmp;  
}

int _db_open(char *f,int mode,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to open is NULL")
  if ((tmp=open(f,mode))>=0)
    _db_add_filedescriptor(tmp,_FILE)
  else
    _db_perror("open")
  return tmp;
}

int _db_putc(int chr,FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("putc on unknown file")
  return putc(chr,f);
}

int _db_read(int f,char *buf,unsigned int cnt,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE|_SOCKET))
    PE("read on unknown file")
  if (buf==NULL)
    PE("argument 2 to read is NULL")
  if (cnt<0)
    PE("argument 3 to read < 0")
  if ((tmp=read(f,buf,cnt))==-1)
    _db_perror("read")
  _db_checksize(buf,cnt,"read",2);
  return tmp;
}

int _db_remove(char *f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to remove is NULL")
  if ((tmp=remove(f))==-1)
    _db_perror("remove")
  return tmp;
}

int _db_rename(char *f,char *tar,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to rename is NULL")
  if (tar==NULL)
    PE("argument 2 to rename is NULL")
  if ((tmp=rename(f,tar))==-1)
    _db_perror("rename")
  return tmp;
}

void _db_rewind(FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE))
    PE("rewind on unknown file")
  rewind(f);
}

void _db_setbuf(FILE *f,char *buf,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("setbuf on unknown file")
  setbuf(f,buf);
}

int _db_setvbuf(FILE *f,char *b,int m,int s,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("setvbuf on unknown file")
  if (s && b==NULL)
    PE("argument 2 to setvbuf is NULL while argument 4 > 0")
  if (s<0)
    PE("argument 4 to setvbuf < 0")
  if (m!=_IOFBF && m!=_IONBF && m!=_IOLBF)
    PE("argument 3 to setvbuf has invalid value")
  return setvbuf(f,b,m,s);
}

char *_db_tmpnam(char *f,char *file,char *function,int line)
{
  char *tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to tmpnam is NULL")
  if ((tmp=tmpnam(f))==NULL)
    PE("tmpnam returned NULL")
  return tmp;
}

FILE *_db_tmpfile(char *file,char *function,int line)
{
  FILE *tmp;
  
  _db_initialize;
  tmp=tmpfile();
  if (tmp!=NULL)
    _db_add_filepointer(tmp,_FILE)
  else
    _db_perror("tmpfile")
  return tmp;
}

int _db_ungetc(int chr,FILE *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("ungetc on unknown file")
  return ungetc(chr,f);
}

int _db_unlink(char *f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to unlink is NULL")
  if ((tmp=unlink(f))==-1)
    _db_perror("unlink")
  return tmp;
}

int _db_write(int f,char *b,unsigned int n,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE|_SOCKET))
    PE("write on unknown file")
  if (b==NULL)
    PE("argument 2 to write is NULL")
  if (n<0)
    PE("argument 3 to write < 0")
  _db_checksize(b,n,"write",2)
  if ((tmp=write(f,b,n))==-1)
    _db_perror("write")
  return tmp;
}

int _db_pipe(int files[2],char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if ((tmp=pipe(files))!=-1)
  {
    _db_add_filedescriptor(files[0],_FILE)
    _db_add_filedescriptor(files[1],_FILE)
  }
  else
    _db_perror("pipe")
  return tmp;
}

int _db_dup(int f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE|_SOCKET))
    PE("dup of unknown file")
  if ((tmp=dup(f))==-1)
    _db_perror("dup")
  else
    _db_add_filedescriptor(tmp,_FILE);
  return tmp;
}

int _db_dup2(int f,int f2,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE|_SOCKET))
    PE("dup2 of unknown file")
  if ((tmp=dup2(f,f2))==-1)
    _db_perror("dup2")
  else
  {
    _db_delete_filedescriptor(f2,_FILE|_SOCKET);
    _db_add_filedescriptor(tmp,_FILE);
  }
  return tmp;
}

FILE *_db_fdopen(int f,char *mode,char *file,char *function,int line)
{
  FILE *tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE|_SOCKET))
    PE("fdopen of unknown file")
  if ((tmp=fdopen(f,mode))==NULL)
    _db_perror("fdopen")
  else
    _db_add_filepointer(tmp,_FILE);
  return tmp;
}

int _db_fstat(int f,struct stat *st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE|_SOCKET))
    PE("fstat of unknown file")
  if ((tmp=fstat(f,st))==-1)
    _db_perror("fstat")
  return tmp;
}

int _db_lstat(char *f,struct stat *st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to lstst is NULL")
  if ((tmp=lstat(f,st))==-1)
    _db_perror("lstat")
  return tmp;
}

int _db_stat(char *f,struct stat *st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to stst is NULL")
  if ((tmp=stat(f,st))==-1)
    _db_perror("stat")
  return tmp;
}

int _db_access(char *f,int st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to access is NULL")
  if ((tmp=access(f,st))==-1)
    _db_perror("access")
  return tmp;
}

DIR *_db_opendir(char *f,char *file,char *function,int line)
{
  DIR *tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to opendir is NULL")
  if ((tmp=opendir(f))==NULL)
    _db_perror("opendir")
  else
    _db_add_filepointer((FILE*)tmp,_DIR);
  return tmp;
}

struct dirent *_db_readdir(DIR *f,char *file,char *function,int line)
{
  _db_initialize;
  if (!_db_check_filepointer((FILE*)f,_DIR))
    PE("readdir of unknown directory")
  return readdir(f);
}

int _db_closedir(DIR *f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_delete_filepointer((FILE*)f,_DIR))
    PE("closedir of unknown directory")
  if ((tmp=closedir(f))==-1)
    _db_perror("closedir")
  return tmp;
}

int _db_chmod(char *f,mode_t st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to chmod is NULL")
  if ((tmp=chmod(f,st))==-1)
    _db_perror("chmod")
  return tmp;
}

int _db_fchmod(int f,mode_t st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE))
    PE("fchmod of unknown file")
  if ((tmp=fchmod(f,st))==-1)
    _db_perror("fchmod")
  return tmp;
}

int _db_chown(char *f,uid_t own,gid_t grp,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to chown is NULL")
  if ((tmp=chown(f,own,grp))==-1)
    _db_perror("chown")
  return tmp;
}

int _db_fchown(int f,uid_t own,gid_t grp,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE))
    PE("fchown of unknown file")
  if ((tmp=fchown(f,own,grp))==-1)
    _db_perror("fchown")
  return tmp;
}

int _db_chdir(char *f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to chdir is NULL")
  if ((tmp=chdir(f))==-1)
    _db_perror("chdir")
  return tmp;
}

int _db_fchdir(int f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE))
    PE("fchdir of unknown file")
  if ((tmp=fchdir(f))==-1)
    _db_perror("fchdir")
  return tmp;
}

int _db_rmdir(char *f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to rmdir is NULL")
  if ((tmp=rmdir(f))==-1)
    _db_perror("rmdir")
  return tmp;
}

int _db_truncate(char *f,size_t st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to truncate is NULL")
  if ((tmp=truncate(f,st))==-1)
    _db_perror("truncate")
  return tmp;
}

int _db_ftruncate(int f,size_t st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_FILE))
    PE("ftruncate of unknown file")
  if ((tmp=ftruncate(f,st))==-1)
    _db_perror("ftruncate")
  return tmp;
}

int _db_mkdir(char *f,mode_t st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to mkdir is NULL")
  if ((tmp=mkdir(f,st))==-1)
    _db_perror("mkdir")
  return tmp;
}

int _db_link(char *f,char *st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to link is NULL")
  if (st==NULL)
    PE("argument 2 to link is NULL")
  if ((tmp=link(f,st))==-1)
    _db_perror("link")
  return tmp;
}

int _db_select(int n,int *r,int *w,int *e,struct timeval *t,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (n<0)
    PE("argument 1 to select < 0")
  _db_checksize(r,n/(sizeof(int)*8)+1,"select",2)
  _db_checksize(w,n/(sizeof(int)*8)+1,"select",3)
  _db_checksize(e,n/(sizeof(int)*8)+1,"select",4)
  if (r!=NULL)
  {
    for (tmp=0;tmp<n;tmp++)
    {
      if (BITSET(tmp,r))
      {
        if (!_db_check_filedescriptor(tmp,_FILE|_SOCKET))
          PE("unknown file descriptor in read array of select")
      }
    }
  }
  if (w!=NULL)
  {
    for (tmp=0;tmp<n;tmp++)
    {
      if (BITSET(tmp,w))
      {
        if (!_db_check_filedescriptor(tmp,_FILE|_SOCKET))
          PE("unknown file descriptor in write array of select")
      }
    }
  }
  if (e!=NULL)
  {
    for (tmp=0;tmp<n;tmp++)
    {
      if (BITSET(tmp,e))
      {
        if (!_db_check_filedescriptor(tmp,_FILE|_SOCKET))
          PE("unknown file descriptor in exception array of select")
      }
    }
  }
  if ((tmp=select(n,(void*)r,(void*)w,(void*)e,t))==-1)
    _db_perror("mkdir")
  return tmp;
}

FILE *_db_popen(char *cmd,char *st,char *file,char *function,int line)
{
  FILE *tmp;
  
  _db_initialize;
  if (cmd==NULL)
    PE("argument 1 to popen is NULL")
  if (st==NULL)
    PE("argument 2 to popen is NULL")
  if ((tmp=popen(cmd,st))==NULL)
    _db_perror("popen")
  else
    _db_add_filepointer(tmp,_PIPE);
  return tmp;
}

int _db_pclose(FILE *f,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_delete_filepointer(f,_PIPE))
    PE("pclose of unknown pipe")
  if ((tmp=pclose(f))==-1)
    _db_perror("pclose")
  return tmp;
}

int _db_socket(int f,int st,int re,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if ((tmp=socket(f,st,re))==-1)
    _db_perror("socket")
  else
    _db_add_filedescriptor(tmp,_SOCKET);
  return tmp;
}

int _db_bind(int f,struct sockaddr *st,int re,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_SOCKET))
    PE("bind of unknown socket")
  if (st==NULL)
    PE("argument 2 to bind is NULL")
  if ((tmp=bind(f,st,re))==-1)
    _db_perror("bind")
  return tmp;
}

int _db_recv(int f,void *s,int n,unsigned int m,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_SOCKET))
    PE("recv on unknown socket")
  if (s==NULL)
    PE("argument 2 to recv is NULL")
  _db_checksize(s,n,"recv",2)
  if ((tmp=recv(f,s,n,m))==-1)
    _db_perror("recv")
  return tmp;
}

int _db_send(int f,void *s,int n,unsigned int m,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_SOCKET))
    PE("send on unknown socket")
  if (s==NULL)
    PE("argument 2 to send is NULL")
  _db_checksize(s,n,"send",2)
  if ((tmp=send(f,s,n,m))==-1)
    _db_perror("send")
  return tmp;
}

int _db_listen(int f,int st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_SOCKET))
    PE("recv on unknown socket")
  if (st<0)
    PE("argument 2 to listen < 0")
  if ((tmp=listen(f,st))==-1)
    _db_perror("listen")
  return tmp;
}

int _db_shutdown(int f,int st,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_SOCKET))
    PE("shutdown of unknown socket")
  if (st<0 || st>2)
    PE("argument 2 to shutdown is invalid")
  if ((tmp=shutdown(f,st))==-1)
    _db_perror("shutdown")
  return tmp;
}

int _db_accept(int f,struct sockaddr *st,int *re,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_SOCKET))
    PE("accept on unknown socket")
  if ((tmp=accept(f,st,re))==-1)
    _db_perror("accept")
  else
    _db_add_filedescriptor(tmp,_SOCKET)
  return tmp;
}

int _db_connect(int f,struct sockaddr *st,int re,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_SOCKET))
    PE("connect of unknown socket")
  if (st==NULL)
    PE("argument 2 to connect is NULL")
  if ((tmp=connect(f,st,re))==-1)
    _db_perror("connect")
  return tmp;
}

int _db_sendto(int f,void *s,int n,unsigned int m,struct sockaddr *a,int l,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_SOCKET))
    PE("sendto on unknown socket")
  if (s==NULL)
    PE("argument 2 to sendto is NULL")
  if (a==NULL)
    PE("argument 5 to sendto is NULL")
  _db_checksize(s,n,"sendto",2)
  if ((tmp=sendto(f,s,n,m,a,l))==-1)
    _db_perror("sendto")
  return tmp;
}

int _db_recvfrom(int f,void *s,int n,unsigned int m,struct sockaddr *a,int *l,char *file,char *function,int line)
{
  int tmp;
  
  _db_initialize;
  if (!_db_check_filedescriptor(f,_SOCKET))
    PE("recvfrom on unknown socket")
  if (s==NULL)
    PE("argument 2 to recvfrom is NULL")
  if (a==NULL)
    PE("argument 5 to recvfrom is NULL")
  _db_checksize(s,n,"recvfrom",2)
  if ((tmp=recvfrom(f,s,n,m,a,l))==-1)
    _db_perror("recvfrom")
  return tmp;
}

char *_db_getcwd(char *f,size_t st,char *file,char *function,int line) 
{
  char *tmp;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to getcwd is NULL")
  if (st<0)
    PE("argument 2 to getcwd < 0")
  _db_checksize(f,st,"getcwd",1)
  if ((tmp=getcwd(f,st))==NULL)
    _db_perror("getcwd")
  return tmp;
}

int _db_scandir(char *f,struct dirent ***s,int (*n)(const struct dirent*),int (*m)(const struct dirent *,const struct dirent *),char *file,char *function,int line)
{
  int tmp,x;
  
  _db_initialize;
  if (f==NULL)
    PE("argument 1 to scandir is NULL")
  if (n==NULL)
    PE("argument 3 to scandir is NULL")
  if (m==NULL)
    PE("argument 4 to scandir is NULL")
  if ((tmp=scandir(f,s,n,(void*)m))==-1)
    _db_perror("scandir")
  else
  {
    for (x=0;x<tmp;x++)
      _db_add_pointer((*s)[x],sizeof(struct dirent));
    _db_add_pointer(*s,sizeof(struct dirent*)*tmp);
  }
  return tmp;
}

void _db_setfileline(char *file,char *function,int line)
{
  _db_initialize;
  strncpy(_db_lastfunction,function,31);
}

int _db_fprintf(FILE *f,char *str,...)
{
  va_list app;
  int tmp,line=_db_lastline;
  char *file=_db_lastfile,*function=_db_lastfunction;
  
  if (!_db_check_filepointer(f,_FILE|_PIPE))
    PE("fprintf on unknown file")
  va_start(app,str);
  tmp=vfprintf(f,str,app);
  va_end(app);
  return tmp;
}
 
int _db_sprintf(char *s,char *str,...)
{
  va_list app;
  int tmp,line=_db_lastline;
  char *file=_db_lastfile,*function=_db_lastfunction;
  
  if (s==NULL)
    PE("argument 1 to sprintf is NULL")
  if (str==NULL)
    PE("argument 2 to sprintf is NULL")
  _db_checksize(s,strlen(str),"sprintf",1)
  va_start(app,str);
  tmp=vsprintf(s,str,app);
  va_end(app);
  return tmp;
}

void _db_startprocedure(char *file,char *function,int line,char *str)
{
  _db_initialize;
  strncpy(_db_lastfunction,function,31);
  if (!function[0] && str)
    _db_Push(_db_stack,file,str,line);
  else
    _db_Push(_db_stack,file,function,line);
}

void _db_endprocedure(char *file,char *function,int line)
{
  _db_Pop(_db_stack);
}

