/* Hash table definitions */

typedef struct {
		 unsigned entries;
		 unsigned allocated;
		 void **data;
	       } HashTableEntry;

typedef struct {
		 unsigned buckets;
		 unsigned key_size;
		 unsigned rec_size;
		 HashTableEntry *data;
	       } HashTable;

typedef void (*HTForEachProc)(void * record);

/* Creates a Hash Table
 * Parameters:
 *   buckets:  The number of buckets, should be about the expexted 
 *             number of records in the hashtable
 *   key_size: The size (in bytes) of the key value of a record.
 *             The key of a record must be at byte offset 0.
 *             (e.g. the first element of a struct)
 *   
 *   rec_size: The total size of one record (in bytes)    
 * Return value: the hash table
*/
HashTable HTCreate(int buckets,int key_size,int rec_size);

/* Adds a record to a Hash Table
 * Parameters:
 *   ht:     The hash table
 *   record: A pointer to the record to be added
 * Return value: none
 * Remarks: This function does not check for duplicate values!
*/
void HTAddRecord(HashTable ht,void *record);

/* Checks if a certain record exists in the hashtable
 * Parameters:
 *   ht:  The hash table
 *   key: A pointer to a key value
 * Return value: 1 if such a record exists
 *               0 if no such record exists
*/
int HTLookup(HashTable ht,void *key);

/* Returns a certain record
 * Parameters:
 *   ht:     The hash table
 *   key:    A pointer to a key value
 *   record: A pointer to a record, this will be filled by the function
 * Return value: 1 if such a record exists
 *               0 if no such record exists
*/
int HTGetRecord(HashTable ht,void *key,void *record);

/* Deletes a certain record
 * Parameters:
 *   ht:  The hash table
 *   key: A pointer to a key value
 * Return value: 1 if such a record exists
 *               0 if no such record exists
*/
int HTDeleteRecord(HashTable ht,void *key);

/* Destroys a hash table
 * Parameters:
 *   ht:  The hash table
 * Return value: none
*/
void HTDestroy(HashTable ht);

/* Calls a function for each element in the hash table
 * Parameters:
 *   ht:     The hash table
 *   feproc: A function pointer. This function will be called for each
             record in the hashtable
 * Return value: none
*/
void HTForEach(HashTable ht,HTForEachProc feproc);

/* Returns the number of records in the hashtable
 * Parameters:
 *   ht:  The hash table
 * Return value: The number of records in the hashtable
*/
int HTNumberOfRecords(HashTable ht);
