#include <stdio.h>
#include <string.h>
#include <Xm/MessageB.h>
#include <Xm/LabelG.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/TextF.h>
#include <Xm/RowColumn.h>
#include <Xm/SeparatoG.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include "match.h"
#include "jex.h"

extern Widget top,combo1,combo2,combo3;

void AddItem(Widget list,char *newtext)
{
  char *text;
  XmString str,*strlist;
  int u_bound,l_bound=0,i;
  
SP("AddItem");
  XtVaGetValues(list,XmNitemCount,&u_bound,
  	XmNitems,&strlist,
 	NULL);
  u_bound--;
  while (u_bound>=l_bound)
  {
    i=l_bound+(u_bound-l_bound)/2;
    if (!XmStringGetLtoR(strlist[i],XmSTRING_DEFAULT_CHARSET,&text))
      break;
    if (db_strcmp(text,newtext)>0)
      u_bound=i-1;
    else
      l_bound=i+1;
    XtFree(text);
  }
  str=XmStringCreateLocalized(newtext);
  XmListAddItemUnselected(list,str,l_bound+1);
  XmStringFree(str);
EP;
}

void SplitPath(char *orig_path,char *directory,char *filter)
{
  char *tmp;
  char path[255];
  
SP("SplitPath");
  db_strcpy(path,orig_path);
  tmp=db_strrchr(path,'/');
  if (tmp)
  {
    *tmp=0;
    db_strcpy(directory,path);
    if (!directory[0])
      db_strcpy(directory,"/");
    db_strcpy(filter,tmp+1);
    if (!filter[0])
      db_strcpy(filter,"*");
  }
  else
  {
    db_strcpy(directory,".");
    db_strcpy(filter,path);
    if (!filter[0])
      db_strcpy(filter,"*");
  }
EP;
}

void GetDirectory(Editor *e,char *directory,char *filter)
{
  DIR *dp;
  struct dirent *dir;
  struct stat sbuf;
  char file[255];
  
SP("GetDirectory");
  XmListDeleteAllItems(e->filefield);
  XmListDeleteAllItems(e->dir_field);
  if ((dp=db_opendir(directory))==NULL)
  {
/*    fprintf(stderr,"Can't open directory %s\n",path);*/
EP;
    return;
  }
  XtUnmanageChild(e->filefield);
  XtUnmanageChild(e->dir_field);
  while ((dir=db_readdir(dp))!=NULL)
  {
    if (dir->d_ino==0)
      continue;
    if (directory[db_strlen(directory)-1]!='/')
      sprintf(file,"%s/%s",directory,dir->d_name);
    else
      sprintf(file,"%s%s",directory,dir->d_name);
    if (db_stat(file,&sbuf)<0)
    {
      if (db_lstat(file,&sbuf)<0)
      {
        continue;
      }
    }
    if ((sbuf.st_mode&S_IFMT)==S_IFDIR)
    {
      AddItem(e->dir_field,file);
      continue;
    }
    if (match(filter,dir->d_name))
    {
      AddItem(e->filefield,dir->d_name);
    }
  }
  db_closedir(dp);
  XtManageChild(e->filefield);
  XtManageChild(e->dir_field);
EP;
}

void SelectDir(Widget w,Editor *e,XmListCallbackStruct *cbs)
{
  char *newdir,*curdir,directory[255],*filter,newpath[255],prevdir[255],tmp[255];
  char olddir[255],newdirectory[255];
  XmString size;
  
SP("SelectDir");
  WatchCursor(e,1);
  XmStringGetLtoR(cbs->item,XmSTRING_DEFAULT_CHARSET,&newdir);
  XtVaGetValues(e->dirfilter_field,XmNvalue,&curdir,NULL);
  XtVaGetValues(e->filter_field,XmNvalue,&filter,NULL);
  SplitPath(newdir,olddir,newdirectory);
  if (db_strcmp(newdirectory,".")==0)
  {
    if (olddir[db_strlen(prevdir)-1]!='/')
      sprintf(newpath,"%s/%s",olddir,filter);
    else
      sprintf(newpath,"%s%s",olddir,filter);
  }
  else if (db_strcmp(newdirectory,"..")==0)
  {
    SplitPath(curdir,prevdir,tmp);
    if (prevdir[db_strlen(prevdir)-1]!='/')
      sprintf(newpath,"%s/%s",prevdir,filter);
    else
      sprintf(newpath,"%s%s",prevdir,filter);
  }
  else
  {
    if (newdir[db_strlen(newdir)-1]!='/')
      sprintf(newpath,"%s/%s",newdir,filter);
    else
      sprintf(newpath,"%s%s",newdir,filter);
  } 
  SplitPath(newpath,directory,filter);
  XtVaSetValues(e->dirfilter_field,XmNvalue,directory,NULL);
  XtVaSetValues(e->sel_field,XmNvalue,directory,NULL);
  size=XmStringCreateLocalized("0 bytes");
  XtVaSetValues(e->fssize,XmNlabelString,size,NULL);
  GetDirectory(e,directory,filter);
  XmStringFree(size);
  XtFree(curdir);
  XtFree(filter);
  XtFree(newdir);
  WatchCursor(e,0);
EP;
}
 
void BrowseSelect(Widget w,Editor *e,XmListCallbackStruct *cbs)
{
  char *newfile,*curdir,filter[255];
  struct stat st;
  XmString size;
  
SP("BrowseSelect");
  XmStringGetLtoR(cbs->item,XmSTRING_DEFAULT_CHARSET,&newfile);
  XtVaGetValues(e->dirfilter_field,XmNvalue,&curdir,NULL);
  if (curdir[db_strlen(curdir)-1]!='/')
    sprintf(filter,"%s/%s",curdir,newfile);
  else
    sprintf(filter,"%s%s",curdir,newfile);
  XtVaSetValues(e->sel_field,XmNvalue,filter,NULL);
  db_stat(filter,&st);
  if (st.st_size>1024*1024)
    sprintf(filter,"%.1f Mb",(double)st.st_size/(1024.0*1024.0));
  else if (st.st_size>1024)
    sprintf(filter,"%.1f Kb",(double)st.st_size/1024.0);
  else
    sprintf(filter,"%lu bytes",(unsigned long)st.st_size);
  size=XmStringCreateLocalized(filter);
  XtVaSetValues(e->fssize,XmNlabelString,size,NULL);
  XmStringFree(size);
  XtFree(newfile);
  XtFree(curdir);
EP;
}

void CallFunction(void (*f)(),Editor *e)
{
SP("CallFunction");
  f(e->fs,e,e->sel_field);
EP;
}

void FSOKCallback(Widget w,Editor *e,void *cbs)
{
  char *str;
  int x;
  
SP("FSOKCallback");
  XtVaGetValues(e->sel_field,XmNvalue,&str,NULL);
  UpdateComboBox(combo3,str);
  for (x=0;x<editors;x++)
  {
    if (editor[x]->fs)
      UpdateComboBox(editor[x]->sel_field,str);
  }
  XtFree(str);
  XtUnmanageChild(e->fs);
  if (e->func!=NULL)
    CallFunction((void(*)(void))e->func,e);
EP;
}
 
void FSCancelCallback(Widget w,Editor *e,void *cbs)
{
SP("FSCancelCallback");
  XtUnmanageChild(e->fs);
EP;
}
 
void FSFilterCallback(Widget w,Editor *e,void *cbs)
{
  char *filter,*dir;
  XmString size;
  int x;
  
SP("FSFilterCallback");
  XtVaGetValues(e->filter_field,XmNvalue,&filter,NULL);
  XtVaGetValues(e->dirfilter_field,XmNvalue,&dir,NULL);
  UpdateComboBox(combo1,filter);
  UpdateComboBox(combo2,dir);
  for (x=0;x<editors;x++)
  { 
    UpdateComboBox(editor[x]->filter_field,filter);
    UpdateComboBox(editor[x]->dirfilter_field,dir);
  }
  GetDirectory(e,dir,filter);
  size=XmStringCreateLocalized("0 bytes");
  XtVaSetValues(e->fssize,XmNlabelString,size,NULL);
  XmStringFree(size);
  XtFree(dir);
  XtFree(filter);
EP;
}
 
void CreateFileSelectionDialog(Editor *e,char *title,void *f,void *f2)
{
  XmString xstr;
  Arg arg[10];
  char path[255];
  Widget w,rc,l;
  int n;
  
SP("CreateFileSelectionDialog");
  e->func=f;
  e->func2=f2;
  WatchCursor(e,1);
  if (e->fs==NULL)
  {
    XtSetArg(arg[0],XmNautoUnmanage,False);
    XtSetArg(arg[1],XmNdefaultButtonType,NULL);
    e->fs=XmCreateMessageDialog(e->shell,"FileSelection",arg,2);
    XtAddCallback(e->fs,XmNokCallback,(XtCallbackProc)FSOKCallback,(void*)e);
    XtAddCallback(e->fs,XmNcancelCallback,(XtCallbackProc)FSCancelCallback,(void*)e);
    XtAddCallback(e->fs,XmNhelpCallback,(XtCallbackProc)FSFilterCallback,(void*)e);
    xstr=XmStringCreateLocalized("Filter");
    XtVaSetValues(e->fs,
    	XmNhelpLabelString,xstr,
    	XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL,
    	NULL);
    XmStringFree(xstr);
    e->form=XtVaCreateManagedWidget("FSForm",xmFormWidgetClass,e->fs,NULL);
    l=XtVaCreateManagedWidget("FSTitle",xmLabelGadgetClass,e->form,
    	XmNtopAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_FORM,
    	XmNleftAttachment,XmATTACH_FORM,
    	NULL);
    l=XtVaCreateManagedWidget("FSTitle",xmSeparatorGadgetClass,e->form,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,l,
    	XmNrightAttachment,XmATTACH_FORM,
    	XmNleftAttachment,XmATTACH_FORM,
    	NULL);
    rc=XtVaCreateManagedWidget("FSRC",xmRowColumnWidgetClass,e->form,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,l,
    	XmNrightAttachment,XmATTACH_FORM,
    	XmNorientation,XmVERTICAL,
    	XmNmarginWidth,0,
    	XmNmarginHeight,0,
    	NULL);
    w=XtVaCreateManagedWidget("FSFilter",xmLabelGadgetClass,rc,
    	XmNalignment,XmALIGNMENT_BEGINNING,
    	NULL);
    e->filter_field=CreateComboBox(rc,"FSFilterField",NULL,0,10);
    InitializeComboBox(e->filter_field,combo1);
    ComboBoxAddCallback(e->filter_field,(XtCallbackProc)FSFilterCallback,(void*)e);
    XtAddCallback(e->filter_field,XmNactivateCallback,(XtCallbackProc)FSFilterCallback,(void*)e);
    rc=XtVaCreateManagedWidget("FSRC",xmRowColumnWidgetClass,e->form,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,l,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_WIDGET,
    	XmNrightWidget,rc,
    	XmNorientation,XmVERTICAL,
    	XmNmarginWidth,0,
    	XmNmarginHeight,0,
    	NULL);
    w=XtVaCreateManagedWidget("FSFilterDirLabel",xmLabelGadgetClass,rc,
    	XmNalignment,XmALIGNMENT_BEGINNING,
    	NULL);
    e->dirfilter_field=CreateComboBox(rc,"FSFilterDirField",NULL,0,10);
    InitializeComboBox(e->dirfilter_field,combo2);
    ComboBoxAddCallback(e->dirfilter_field,(XtCallbackProc)FSFilterCallback,(void*)e);
    XtAddCallback(e->dirfilter_field,XmNactivateCallback,(XtCallbackProc)FSFilterCallback,(void*)e);
    e->dirlabel=XtVaCreateManagedWidget("FSDirLabel",xmLabelGadgetClass,e->form,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,XtParent(XtParent(e->filter_field)),
    	XmNleftAttachment,XmATTACH_FORM,
    	NULL);
    XtSetArg(arg[0],XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE);
    e->dir_field=XmCreateScrolledList(e->form,"FSDirList",arg,1);
    XtVaSetValues(XtParent(e->dir_field),
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->dirlabel,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNscrollBarDisplayPolicy,XmSTATIC,
    	NULL);
    XtVaSetValues(e->dir_field,
    	XmNscrollBarDisplayPolicy,XmSTATIC,
    	XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE,
    	NULL);
    XtAddCallback(e->dir_field,XmNdefaultActionCallback,(XtCallbackProc)SelectDir,(void*)e);
    XtManageChild(e->dir_field);
    e->filelabel=XtVaCreateManagedWidget("FSFileLabel",xmLabelGadgetClass,e->form,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,XtParent(XtParent(e->filter_field)),
    	XmNleftAttachment,XmATTACH_WIDGET,
    	XmNleftWidget,XtParent(e->dir_field),
    	XmNrightAttachment,XmATTACH_FORM,
    	NULL);
    e->filefield=XmCreateScrolledList(e->form,"FSFileList",arg,1);
    XtVaSetValues(XtParent(e->filefield),
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->filelabel,
    	XmNleftAttachment,XmATTACH_WIDGET,
    	XmNleftWidget,XtParent(e->dir_field),
    	XmNrightAttachment,XmATTACH_FORM,
    	XmNscrollBarDisplayPolicy,XmSTATIC,
    	XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE,
    	NULL);
    XtVaSetValues(e->filefield,
    	XmNscrollBarDisplayPolicy,XmSTATIC,
    	XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE,
    	NULL);
   XtAddCallback(e->filefield,XmNbrowseSelectionCallback,(XtCallbackProc)BrowseSelect,(void*)e);
   XtAddCallback(e->filefield,XmNdefaultActionCallback,(XtCallbackProc)FSOKCallback,(void*)e);
   XtManageChild(e->filefield);
   e->fssize=XtVaCreateManagedWidget("FSSize",xmLabelGadgetClass,e->form,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,XtParent(e->dir_field),
    	XmNleftAttachment,XmATTACH_NONE,
    	XmNrightAttachment,XmATTACH_FORM,
    	NULL);
   e->selection=XtVaCreateManagedWidget("FSSelection",xmLabelGadgetClass,e->form,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,XtParent(e->dir_field),
    	XmNleftAttachment,XmATTACH_FORM,
    	NULL);
    n=0;
    XtSetArg(arg[n],XmNtopAttachment,XmATTACH_WIDGET);n++;
    XtSetArg(arg[n],XmNtopWidget,e->selection);n++;
    XtSetArg(arg[n],XmNleftAttachment,XmATTACH_FORM);n++;
    XtSetArg(arg[n],XmNrightAttachment,XmATTACH_FORM);n++;
    e->sel_field=CreateComboBox(e->form,"FSSelectionField",arg,n,10);
    InitializeComboBox(e->sel_field,combo3);
    XtAddCallback(e->sel_field,XmNactivateCallback,(XtCallbackProc)FSOKCallback,(void*)e);
    db_getcwd(path,255);
    XtVaSetValues(e->filter_field,XmNvalue,"*",NULL);
    XtVaSetValues(e->dirfilter_field,XmNvalue,path,NULL);
    XtVaSetValues(e->sel_field,XmNvalue,path,NULL);
    GetDirectory(e,path,"*");
  }
  else
    FSFilterCallback(NULL,e,NULL); /* Refresh */
  XtVaSetValues(XtParent(e->fs),XmNtitle,title,NULL);
  xstr=XmStringCreateLocalized(title);
#ifndef LESSTIFF    
  XtVaSetValues(XtNameToWidget(e->fs,"*FSTitle"),XmNlabelString,xstr,NULL);
#endif
  XmStringFree(xstr);
  XtManageChild(e->fs);
  WatchCursor(e,0);
EP;
}
