#include "jbdbg.h"

#include <stdio.h>
#include <Xm/Xm.h>
#include "list.h"
#include "help.h"
#include "combo.h"

#define MAXEDITORS 100
#define BUFSIZE 255
#define FILELIST 10
#define MAJOR_VERSION 1
#define MINOR_VERSION 3
#define SUB_MINOR_VERSION 7
#define MAXBOOKMARKBUTTONS 20

typedef struct{
		int pid;
                int fd[2],fd2[2],fd3[2];/* Pipes, and process id */
                XtInputId id,id2;       /* File pointer input descriptors */
                char name[100];		/* Executed command */
              } ExecType;

typedef struct {
		 char *undelete;
		 XmTextPosition pos;
		 int insertsize,deletesize;
	       } Undo_T;
	       
typedef struct {
		 Boolean histhome;
		 Boolean setthome;
               } OwnRes;
              
typedef struct {
		 char filename[255];
		 char buf[BUFSIZE];
		 char choice[255];
		 int x,marker;
		 Widget shell,text,undo,redo,rowcol,save,findf,ato;
		 Widget replf,gotofield,pm,mb,compile,make;
		 Widget ats,tmpsource,bookm[MAXBOOKMARKBUTTONS];
		 Widget filelist[FILELIST],bookmarkmenu;
		 Widget fd,grep,gd,gpat,gfiles,gsubd;
		 Widget iod,io_win,iodtext;
		 int saved,oldsource,same_line;
		 Widget fs,clock,filemenu;
		 Widget form,dirfilter_field,filter_field,dir_field;
		 Widget filefield,selection,sel_field,dirlabel;
		 Widget filelabel,abd,fssize;
		 int undosize,lastaction,redosize,blockundo,changes,ce_count;
		 List undolist;
		 List redolist;
		 Widget le,se,bu,casesens;
		 Widget sd,sf,ai,ull,ul,pcl,pc,cc,ccl,mc,mcl,ce,cel,nw;
		 void *f,*func,*func2;
		 ExecType t;
                 char print_command[100],compile_command[100],grepped;
                 char make_command[100],run_command[100],**compiler_error;
		 Boolean auto_indent,new_window,autosave;
                 int undolevels,workcount,autosavetimeout;
                 XtIntervalId tid,tid2,tid3,tid4,tid5,tid6,tid7,tid8;
                 int locked,read_only;
                 Widget std,statln,statch,statchan,statundo,statredo;
                 Widget nw2,asf,ast,err,run,runcl,runc,owi;
              } Editor;


extern int editors;
extern Editor *editor[MAXEDITORS];

void file_cb();
void edit_cb();
void popup_cb(Widget w,int item,XmAnyCallbackStruct *cbs);
void option_cb();
void bookmark_cb();
void tools_cb();
void view_cb();
void help_cb();
void Save(Widget w,Editor *e,void *cbs);
void AskToSave(Editor *e,void *f);
void ValueChanged(Widget w,Editor *e,void *cbs);
void Grep(Widget w,Editor *e,void *cbs);
void New(Editor *e);
void SaveAs(Editor *e);
void Load(Editor *e);
void Revert(Editor *e);
void Insert(Editor *e);
void Print(Editor *e);
void FindRowCol(Widget w,Editor *e,XmTextVerifyCallbackStruct *cbs);
void FindUp(Widget w,Editor *e,void *cbs);
void FindDown(Widget w,Editor *e,void *cbs);
void Replace(Widget w,Editor *e,void *cbs);
void ReplaceAll(Widget w,Editor *e,void *cbs);
void CutText(Widget w,Editor *e,void *cbs);
void CopyText(Widget w,Editor *e,void *cbs);
void PasteText(Widget w,Editor *e,void *cbs);
void ClearSelection(Widget w,Editor *e,void *cbs);
void DeleteText(Widget w,Editor *e,XmTextVerifyCallbackStruct *cbs);
void UndoAction(Widget w,Editor *e,void *cbs);
void RedoAction(Widget w,Editor *e,void *cbs);
void RepeatAction(Widget w,Editor *e,void *cbs);
void GotoLine(Widget w,Editor *e,void *cbs);
void Quit(Widget w,Editor *e,void *cbs);
void PopupMenu(Widget w,XButtonPressedEvent *event,String *args,int *num_args);
void CreateFileSelectionDialog(Editor *e,char *title,void *f,void *f2);
void CapsSelection(Widget w,Editor *e,void *cbs);
void LowsSelection(Widget w,Editor *e,void *cbs);
void ShiftRight(Widget w,Editor *e,void *cbs);
void ShiftLeft(Widget w,Editor *e,void *cbs);
void Compile(Widget w,Editor *e,void *cbs);
void Make(Widget w,Editor *e,void *cbs);
void UnlinkFile(Editor *e);
void StartEditor(void);
void NoUndo(Editor *e);
void Exit(Editor *e);
void InsertFile(Editor *e,char *str);
int loadfile(Editor *e,char *file,int insert,int setname);
int savefile(Editor *e,char *file,char *str);
void CopyFile(char *s,char *t);
int FindTextString(Editor *e,char *str,int start_over,int set_pos);
int FindReverseTextString(Editor *e,char *str,int start_over);
void ReplaceText(Editor *e,char *str);
void printfile(Editor *e,char *file);
void Terminate(int sig);
void GotoError(Widget w,Editor *e,void *cbs);
void LoadSettings(Editor *e);
void SetTitle(Editor *e,char *str);
void CreateNewEditor(void);
void Settings(Editor *e);
void SplitPath(char *path,char *dir,char *file);
void SaveError(Editor *e);
void LoadError(Editor *e);
void Function(void (*f)(),Editor *e);
void CreateEditorDialog(void);
void ShowEditorDialog(void);
void UpdateEditorDialog(void);
void CreateOutputWindow(Editor *e);
void WatchCursor(Editor *e,int turn_on);
void ShowAboutDialog(Editor *e);
void ExecProg(Editor *e,char *_prog);
void CreateOutputWindow(Editor *e);
void SplitUp(char *str,char *list[]);
void wprintf(Widget window,...);
void SetLabel(Editor *e,...);
void DoGrep(char *pat,char *replace);
void SetWorkingLabel(Editor *e,void *t);
void InputCallback1(Editor *e,int f,XtInputId l);
void InputCallback2(Editor *e,int f,XtInputId l);
void ChildSignal(int sig);
void UpdateFileList(char *file);
void UpdateFindList(char *file);
void UpdateReplList(char *file);
void UpdateGrepList(char *file);
void UpdateGrepFileList(char *file);
void OpenOldFile(Widget w,Editor *e,void *cbs);
void RemoveInputs(Editor *e,void *tid);
void ShowFindMenu(Widget w,Editor *e,XmPushButtonCallbackStruct *cbs);
void ShowReplaceMenu(Widget w,Editor *e,XmPushButtonCallbackStruct *cbs);
void ShowGrepMenu(Widget w,Editor *e,XmPushButtonCallbackStruct *cbs);
void ShowGrepFileMenu(Widget w,Editor *e,XmPushButtonCallbackStruct *cbs);
void SelectFindString(Widget w,Editor *e,void *cbs);
void SelectReplString(Widget w,Editor *e,void *cbs);
void SelectGrepString(Widget w,Editor *e,void *cbs);
void SelectGrepFileString(Widget w,Editor *e,void *cbs);
void GetFullFileName(char *file);
void ShowBookmarkDialog(void);
void AddBookmark(char *str);
void CreateBookmarkDialog(void);
void ShowStats(Editor *e);
void DeleteLine(Widget w,XAnyEvent *event,String *args,int *num_args);
void NoHighlight(Editor *e,void *id);
void FindMatchingChar(Widget w,Editor *e,void *cbs);
void Reload(Editor *e);
void DetachEditor(Editor *e,char *file);
void SetMarker(Widget w,Editor *e,void *cbs);
void GotoMarker(Widget w,Editor *e,void *cbs);
void RemoveMarker(Editor *e);
void SetCurrentWidget(Widget w);
void DoNew(Widget *w,Editor *e,void *cbs);
void DoOpen(Widget *w,Editor *e,void *cbs);
void SetSave(Editor *e,Boolean enabled);
void SmallHelp(Widget w,int n,XEvent *event);
void RemoveSmallHelp(Widget w,int n,void *cbs);
void DoNewWindow(Widget *w,void *e,void *cbs);
void DoClone(Widget *w,Editor *e,void *cbs);
void LatestVersion(void);
void HelpError(char *msg);
int FindHelpFile(char *helpfile,char *file);
void Run(Widget w,Editor *e,void *cbs);
void ShowError(Editor *e,char *errstr);
void CheckVersion(char *version);
void ShowVersion(void);
void LoadOldFile(Editor *e);
void UpdateEditorBookmarks(Editor *e);

