/* LIST.C
   Author     : Jorgen Bosman.
   Last update: 13-04-1994
*/

#include "list.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "jbdbg.h"

#define debug 0

#if debug==1
#define Debug(l,s) VisualizeList(l,s)
#else
#define Debug(l,s)
#endif

typedef struct ListEl {
		 void *element;
		 struct ListEl *next;
		 struct ListEl *prev;
	       } ListElem;

typedef struct {
		 unsigned el_size;
		 ListElem *first;
		 ListElem *last;
		 ListElem *current;
	       } ListDef;

typedef ListDef *_List;

void VisualizeList(List l,char *str)
{
  ListElem *ptr,*ptr2;
  int x=0;

  printf("\n%s:\n  ",str);
  if (!l || !((_List)l)->first)
    printf("List empty");
  else
  {
    ptr=ptr2=((_List)l)->first;
    printf("F");
    if (ptr==((_List)l)->current)
      printf("C");
    if (ptr==((_List)l)->last)
      printf("L");
    printf("%d",*(int*)ptr->element);
    while (ptr->next)
    {
      x++;
      ptr=ptr->next;
      if (ptr->prev==ptr2)
	printf("<-");
      if (ptr2->next==ptr)
	printf("->");
      if (ptr==((_List)l)->current)
	printf("C");
      if (ptr==((_List)l)->last)
	printf("L");
      printf("%d",*(int*)ptr->element);
      ptr2=ptr;
    }
  }
printf("\n");
}

List CreateList(unsigned el_size)
{
  _List l;

SP("CreateList");
  if ((l=(ListDef *)db_malloc(sizeof(ListDef)))!=NULL)
  {
    l->first=NULL;
    l->last=NULL;
    l->current=NULL;
    l->el_size=el_size;
  }
  Debug(l,"Createlist");
EP;
  return (List)l;
}


int ListEmpty(List list)
{
  Debug(list,"ListEmpty");
  return (list==NULL || ((_List)list)->first==NULL);
}

int AddElement(List l,void *element)
{
SP("AddElement");
  Debug(l,"AddElement start");
  if (!l)
  {
    EP;
    return 0;
  }
  if (NULL==((_List)l)->last)
  {
    if ((((_List)l)->last=(ListElem*)db_malloc(sizeof(ListElem)))==NULL)
    {
      EP;
      return 0;
    }
    ((_List)l)->first=((_List)l)->last;
    ((_List)l)->last->prev=NULL;
  }
  else
  {
    if ((((_List)l)->last->next=(ListElem*)db_malloc(sizeof(ListElem)))==NULL)
    {
      EP;
      return 0;
    }
    ((_List)l)->last->next->prev=((_List)l)->last;
    ((_List)l)->last=((_List)l)->last->next;
  }
  if ((((_List)l)->last->element=db_malloc(((_List)l)->el_size))==NULL)
  {
    EP;
    return 0;
  }
  ((_List)l)->last->element=db_memcpy(((_List)l)->last->element,element,((_List)l)->el_size);
  ((_List)l)->last->next=NULL;
  ((_List)l)->current=((_List)l)->last;
  Debug(l,"AddElement end");
  EP;
  return 1;
}

int AddFirstElement(List l,void *element)
{
SP("AddFirstElement");
  Debug(l,"AddFirstElement start");
  if (!l)
  {
    EP;
    return 0;
  }
  if (!((_List)l)->last) /* List is empty. */
  {
    if ((((_List)l)->last=(ListElem*)db_malloc(sizeof(ListElem)))==NULL)
    {
      EP;
      return 0;
    }
    ((_List)l)->first=((_List)l)->last;
    ((_List)l)->last->prev=NULL;
    ((_List)l)->last->next=NULL;
    ((_List)l)->current=((_List)l)->last;
  }
  else
  {
    if ((((_List)l)->first->prev=(ListElem*)db_malloc(sizeof(ListElem)))==NULL)
    {
      EP;
      return 0;
    }
    ((_List)l)->current=((_List)l)->first->prev;
    ((_List)l)->current->next=((_List)l)->first;
    ((_List)l)->current->prev=NULL;
    ((_List)l)->first=((_List)l)->current;
  }
  if ((((_List)l)->first->element=db_malloc(((_List)l)->el_size))==NULL)
  {
    EP;
    return 0;
  }
  ((_List)l)->first->element=db_memcpy(((_List)l)->first->element,element,((_List)l)->el_size);
  Debug(l,"AddFirstElement end");
  EP;
  return 1;
}

int InsertElement(List list,void *element)
{
  ListElem *l;

SP("InsertElement");
  Debug(list,"InsertElement start");
  if (!list || !((_List)list)->first || !((_List)list)->first->prev)
  {
    EP;
    return AddFirstElement(list,element);
  }
  else if (((_List)list)->current==((_List)list)->last)
  {
    EP;
    return AddElement(list,element);
  }
  else
  {
    if (!(l=(ListElem*)db_malloc(sizeof(ListElem))))
    {
      EP;
      return 0;
    }
    l->prev=((_List)list)->current->prev;
    l->next=((_List)list)->current;
    ((_List)list)->current->prev=l;
    ((_List)list)->current->prev->next=l;
    ((_List)list)->current=l;
    if (!(l->element=db_malloc(((_List)list)->el_size)))
    {
      EP;
      return 0;
    }
    l->element=db_memcpy(l->element,element,((_List)list)->el_size);
  }
  Debug(list,"InsertElement Last");
EP;
  return 1;
}


int GetFirstElement(List list,void *element)
{
SP("GetFirstElement");
  Debug(list,"GetFirstElement start");
  if (!list || !((_List)list)->first)
  {
    EP;
    return 0;
  }
  element=db_memcpy(element,((_List)list)->first->element,((_List)list)->el_size);
  ((_List)list)->current=((_List)list)->first;
  Debug(list,"GetFirstElement end");
  EP;
  return 1;
}

int GetCurrentElement(List list,void *element)
{
SP("GetCurrentElement");
  Debug(list,"GetCurrentElement start");
  if (!list || !((_List)list)->current)
  {
    EP;
    return 0;
  }
  element=db_memcpy(element,((_List)list)->current->element,((_List)list)->el_size);
  Debug(list,"GetCurrentElement end");
  EP;
  return 1;
}

int GetNextElement(List list,void *element)
{
SP("GetNextElement");
  Debug(list,"GetNextElement start");
  if (!list || !((_List)list)->current || !((_List)list)->current->next)
  {
    EP;
    return 0;
  }
  ((_List)list)->current=((_List)list)->current->next;
  element=db_memcpy(element,((_List)list)->current->element,((_List)list)->el_size);
  Debug(list,"GetNextElement end");
  EP;
  return 1;
}

int GetPreviousElement(List list,void *element)
{
SP("GetPreviousElement");
  Debug(list,"GetPreviousElement start");
  if (!list || !((_List)list)->current || !((_List)list)->current->prev)
  {
    EP;
    return 0;
  }
  ((_List)list)->current=((_List)list)->current->prev;
  element=db_memcpy(element,((_List)list)->current->element,((_List)list)->el_size);
  Debug(list,"GetPreviousElement end");
  EP;
  return 1;
}

int GetLastElement(List list,void *element)
{
  SP("GetLastElement");
  Debug(list,"GetLastElement start");
  if (!list || !((_List)list)->last)
  {
    EP;
    return 0;
  }
  element=db_memcpy(element,((_List)list)->last->element,((_List)list)->el_size);
  ((_List)list)->current=((_List)list)->last;
  Debug(list,"GetLastElement end");
EP;
  return 1;
}

int GetNthElement(List list,int n,void *element)
{
  int x=0;  

SP("GetNthElement");
  Debug(list,"GetNthElement start");
  if (!list || !((_List)list)->last)
  {
    EP;
    return 0;
  }
  if (GetFirstElement(list,element))
  {
    do
    {
      if (x++==n)
      {
        EP;
	return 1;
      }
    } while (GetNextElement(list,element));
  }      
  Debug(list,"GetNthElement end");
EP;
  return 0;
}

int DisposeList(List list)
{
  ListElem *ptr,*ptr2;

SP("DisposeList");
  Debug(list,"DisposeList start");
  if (!list)
  {
    EP;
    return 0;
  }
  ptr=((_List)list)->first;
  while (ptr)
  {
    ptr2=ptr;
    ptr=ptr->next;
    db_free((char*)ptr2->element);
    db_free((char*)ptr2);
  }
  db_free((char*)list);
  list=NULL;
  Debug(list,"DisposeList end");
  EP;
  return 0;
}

int ClearList(List list)
{
  ListElem *ptr,*ptr2;

SP("ClearList");
  Debug(list,"ClearList start");
  if (!list)
  {
    EP;
    return 0;
  }
  ptr=((_List)list)->first;
  while (ptr)
  {
    ptr2=ptr;
    ptr=ptr->next;
    db_free((char*)ptr2->element);
    db_free((char*)ptr2);
  }
  ((_List)list)->first=NULL;
  ((_List)list)->last=NULL;
  ((_List)list)->current=NULL;
  Debug(list,"ClearList end");
  EP;
  return 0;
}

int FindElement(List list,void *element,int key_size,void *key)
{
  ListElem *ptr;

SP("FindElement");
  Debug(list,"ClearList start");
  ptr=((_List)list)->first;
  while (db_memcmp(ptr->element,key,key_size)!=0 && ptr->next)
    ptr=ptr->next;
  if (db_memcmp(ptr->element,key,key_size)==0)
  {
    element=db_memcpy(element,ptr->element,((_List)list)->el_size);
    ((_List)list)->current=ptr;
EP;
    return 1;
  }
  Debug(list,"ClearList end");
EP;  
  return 0;
}

int DeleteLastElement(List list)
{
SP("DeleteLastElement");
  Debug(list,"DeleteLastElement start");
  if (!list || !((_List)list)->last)
  {
    EP;
    return 0;
  }
  if (((_List)list)->last->prev)
  {
    ((_List)list)->last=((_List)list)->last->prev;
    db_free((char*)((_List)list)->last->next->element);
    db_free((char*)((_List)list)->last->next);
    ((_List)list)->last->next=NULL;
  }
  else
  {
    ((_List)list)->first=NULL;
    ((_List)list)->current=NULL;
    db_free((char*)((_List)list)->last->element);
    db_free((char*)((_List)list)->last);
    ((_List)list)->last=NULL;
  }
  Debug(list,"DeleteLastElement end");
EP;
  return 1;
}

int DeleteFirstElement(List list)
{
SP("DeleteFirstElement");
  Debug(list,"DeleteFirstElement start");
  if (!list || !((_List)list)->first)
  {
    EP;
    return 0;
  }
  if (((_List)list)->first->next)
  {
    db_free((char*)((_List)list)->first->element);
    ((_List)list)->current=((_List)list)->first->next;
    db_free((char*)((_List)list)->current->prev);
    ((_List)list)->current->prev=NULL;
    ((_List)list)->first=((_List)list)->current;
  }
  else
  {
    ((_List)list)->last=NULL;
    ((_List)list)->current=NULL;
    db_free((char*)((_List)list)->first->element);
    db_free((char*)((_List)list)->first);
    ((_List)list)->first=NULL;
  }
  Debug(list,"DeleteFirstElement end");
EP;
  return 1;
}

int DeleteCurrentElement(List list)
{
  ListElem *ptr;

SP("DeleteCurrentElement");
  Debug(list,"DeleteCurrentElement start");
  if (!list)
  {
    EP;
    return 0;
  }
  ptr=((_List)list)->current;
  if (!ptr)
  {
    EP;
    return 0;
  }
  if (ptr->prev)
    ptr->prev->next=ptr->next;
  else
    ((_List)list)->first=ptr->next;
  if (ptr->next)
  {
    ptr->next->prev=ptr->prev;
    ((_List)list)->current=ptr->next;
  }
  else
    ((_List)list)->last=((_List)list)->current=ptr->prev;
  db_free((char*)ptr->element);
  db_free((char*)ptr);
  Debug(list,"DeleteCurrentElement end");
EP;
  return 1;
}

int DeleteElement(List list,int key_size,void *key)
{
  ListElem *ptr;

SP("DeleteElement");
  Debug(list,"DeleteElement start");
  if (!list)
  {
    EP;
    return 0;
  }
  ptr=((_List)list)->first;
  while (db_memcmp(ptr->element,key,key_size)!=0 && ptr->next)
    ptr=ptr->next;
  if (db_memcmp(ptr->element,key,key_size)==0)
  {
    if (ptr->prev)
      ptr->prev->next=ptr->next;
    else
      ((_List)list)->first=ptr->next;
    if (ptr->next)
    {
      ptr->next->prev=ptr->prev;
      ((_List)list)->current=ptr->next;
    }
    else
      ((_List)list)->last=((_List)list)->current=ptr->prev;
    db_free((char*)ptr->element);
    db_free((char*)ptr);
    Debug(list,"DeleteElement end");
    EP;
    return 1;
  }
  Debug(list,"DeleteElement end");
EP;
  return 0;
}

void ReplaceCurrentElement(List list,void *element)
{
SP("ReplaceCurrentElement");
  Debug(list,"ReplaceCurrentElement start");
  if (list==NULL)
  {
    EP;
    return;
  }
  ((_List)list)->current->element=db_memcpy(((_List)list)->current->element,element,((_List)list)->el_size);
EP;
  Debug(list,"ReplaceCurrentElement end");
}

void SortList(List list,int (*sort_proc)(const void *element1,const void *element2))
{
  char *data,*tmp;
  ListElem *ptr1;
  int n;

SP("SortList");
  Debug(list,"SortList start");
  tmp=data=(char*)db_malloc((n=NumberElements(list))*((_List)list)->el_size);
  ptr1=((_List)list)->first;
  while (ptr1)
  {
    data=(char*)db_memcpy(data,ptr1->element,((_List)list)->el_size);
    data+=((_List)list)->el_size;
    ptr1=ptr1->next;
  }
  ClearList(list);
  data=tmp;
  qsort(data,n,((_List)list)->el_size,sort_proc);
  while (n--)
  {
    AddElement(list,data);
    data+=((_List)list)->el_size;
  }
  db_free((char*)tmp);
  Debug(list,"SortList end");
EP;
}

int NumberElements(List list)
{
  ListElem *ptr;
  int count=0;

SP("NumberElements");
  Debug(list,"NumberElements start");
  if (!list)
  {
    EP;
    return 0;
  }
  ptr=((_List)list)->first;
  while (ptr)
  {
    ptr=ptr->next;
    count++;
  }
  Debug(list,"NumberElements end");
EP;
  return count;
}

