/* LIST.H
   Author     : Jorgen Bosman.
   Last Update: 13-04-1994
*/

#ifndef _LIST_H_
#define _LIST_H_

typedef void *List;

List CreateList(unsigned el_size);
/* Creates new definition list.
   Must be called before any other function. */

int AddElement(List list,void *element);
/* Adds an element to back of the list. */

int AddFirstElement(List list,void *element);
/* Adds an element to the front of the list. */

int InsertElement(List list,void *element);
/* Insert the element in the current position. */

int GetFirstElement(List list,void *element);

int GetElement(List list,void *element);

int GetNextElement(List list,void *element);

int GetPreviousElement(List list,void *element);

int GetLastElement(List list,void *element);

int GetNthElement(List list,int n,void *element);
/* Gets the n-th element from the list starting from 0.
   If the list is smaller than n, element contains the last element and 0 is
   returned, otherwise 0 is returned. */

int FindElement(List list,void *element,int key_size,void *key);
/* Finds element in list using key and the size of the key,
   the key must be at the beginning of the element record. */

int DeleteLastElement(List list);
/* Deletes last element from list. */

int DeleteFirstElement(List list);
/* Deletes first element from list. */

int DeleteCurrentElement(List l);
/* Deletes current element from list. */

int DeleteElement(List list,int key_size,void *key);
/* Deletes element in list using key and the size of the key,
   the key must be at the beginning of the element record. */

int ListEmpty(List list);
/* Checks if list is empty. */

int DisposeList(List list);
/* Disposes list. */

int ClearList(List list);
/* Clears list. */

void ReplaceCurrentElement(List list,void *element);
/* Replaces current element by element. */

void SortList(List list,int (*sort_proc)(const void *element1,const void *element2));
/* Sorts list using compare function sort_proc(), it must return 1
   if element1 > element2, 0 otherwise. */

int NumberElements(List l);
/* Returns number of elements in list. */

void VisualizeList(List l,char *str);
/* Makes list visual on screen, it displays each element as integer.
   Str is printed before the list.
   Symbols: F: first element, C: current element, L:last element */

#endif

