#include <stdio.h>
#include <stdlib.h>
#include <Xm/MessageB.h>
#include <Xm/Form.h>
#include <Xm/TextF.h>
#include <Xm/LabelG.h>
#include <Xm/ToggleB.h>
#include <Xm/PushBG.h>
#include <Xm/List.h>
#include <Xm/RowColumn.h>
#include "jex.h"

extern Widget top;
extern OwnRes ownres;
extern XtAppContext app;

String auto_indent_on="#override \
Ctrl<Key>Y: delete-line() \n\
<Key>Return: newline-and-indent()";

String auto_indent_off="#override \
Ctrl<Key>Y: delete-line() \n\
<Key>Return: newline()";

void AutoSave(Editor *e,XtIntervalId id)
{
SP("AutoSave");
  if (e->filename[0])
    Save(NULL,e,NULL);
  e->tid4=XtAppAddTimeOut(app,e->autosavetimeout*60000,(XtTimerCallbackProc)AutoSave,e);
EP;
}     

void ToggleAutoSave(Widget w,Editor *e,XmToggleButtonCallbackStruct *cbs)
{
SP("ToggleAutoSave");
  XtVaSetValues(e->asf,XmNsensitive,cbs->set?True:False,NULL);
EP;
}

void ApplySettings(Editor *e)
{
SP("ApplySettings");
  if (e->auto_indent)
    XtOverrideTranslations(e->text,XtParseTranslationTable(auto_indent_on));
  else
    XtOverrideTranslations(e->text,XtParseTranslationTable(auto_indent_off));
  if (e->autosave)
  {
    if (e->sd)
      XtVaSetValues(e->asf,XmNsensitive,True,NULL);
    if (e->tid4)
      XtRemoveTimeOut(e->tid4);
    e->tid4=XtAppAddTimeOut(app,e->autosavetimeout*60000,(XtTimerCallbackProc)AutoSave,e);
  }
  else
  {
    if (e->sd)
      XtVaSetValues(e->asf,XmNsensitive,False,NULL);
    if (e->tid4)
      XtRemoveTimeOut(e->tid4);
  }
EP;
}

void SaveSettings(Editor *e)
{
  FILE *f;
  char filename[255],*home;
  int x;
  
SP("SaveSettings");
  if (ownres.setthome)
  {
    home=getenv("HOME");
    if (home)
      sprintf(filename,"%s/.jexrc",home);
    else
      db_strcpy(filename,".jexrc");
  }
  else
    db_strcpy(filename,".jexrc");
  if ((f=db_fopen(filename,"w+"))==NULL)
  {
    EP;
    return;
  }
  fprintf(f,"autoindent: %d\n",(int)e->auto_indent);
  fprintf(f,"undolevels: %d\n",e->undolevels);
  fprintf(f,"printcommand: %s\n",e->print_command);
  fprintf(f,"compilecommand: %s\n",e->compile_command);
  fprintf(f,"makecommand: %s\n",e->make_command);
  fprintf(f,"runcommand: %s\n",e->run_command);
  for (x=0;x<e->ce_count;x++)
    fprintf(f,"compilererror: %s\n",e->compiler_error[x]);
  fprintf(f,"newwindow: %d\n",(int)e->new_window);
  fprintf(f,"blockundo: %d\n",(int)e->blockundo);
  fprintf(f,"autosave: %d\n",(int)e->autosave);
  fprintf(f,"autosavetimeout: %d\n",(int)e->autosavetimeout);
  db_fclose(f);
EP;
}

void LoadSettings(Editor *e)
{
  FILE *f;
  char filename[255],*home;
  int tmp,x;
  
SP("LoadSettings");
  if (e->ce_count)
  {
    for (x=0;x<e->ce_count;x++)
      db_free(e->compiler_error[x]);
    db_free(e->compiler_error);
    e->compiler_error=NULL;
    e->ce_count=0;
  }    
  if (ownres.setthome)
  {
    home=getenv("HOME");
    if (home)
      sprintf(filename,"%s/.jexrc",home);
    else
      db_strcpy(filename,".jexrc");
  }
  else
    db_strcpy(filename,".jexrc");
  if ((f=fopen(filename,"r"))==NULL)
  {
    if (!e->ce_count)
    {
      e->compiler_error=(char**)db_realloc(e->compiler_error,(e->ce_count+1)*sizeof(char*));
      e->compiler_error[e->ce_count++]=db_strdup("%[^:]:%d");
      e->compiler_error=(char**)db_realloc(e->compiler_error,(e->ce_count+1)*sizeof(char*));
      e->compiler_error[e->ce_count++]=db_strdup("cc: \"%[^\"]\", line %d:");
    }
    ApplySettings(e);
    EP;
    return;
  }
  while (fscanf(f,"%s",filename)==1)
  {
    if (db_strcmp(filename,"autoindent:")==0)
    {
      fscanf(f,"%d",&tmp);
      e->auto_indent=(Boolean)tmp;
    }
    else if (db_strcmp(filename,"newwindow:")==0)
    {
      fscanf(f,"%d",&tmp);
      e->new_window=(Boolean)tmp;
    }
    else if (db_strcmp(filename,"blockundo:")==0)
    {
      fscanf(f,"%d",&tmp);
      e->blockundo=(Boolean)tmp;
    }
    else if (db_strcmp(filename,"autosave:")==0)
    {
      fscanf(f,"%d",&tmp);
      e->autosave=(Boolean)tmp;
    }
    else if (db_strcmp(filename,"autosavetimeout:")==0)
      fscanf(f,"%d",&e->autosavetimeout);
    else if (db_strcmp(filename,"undolevels:")==0)
      fscanf(f,"%d",&e->undolevels);
    else if (db_strcmp(filename,"printcommand:")==0)
    {
      fgets(filename,100,f);
      filename[db_strlen(filename)-1]='\0';
      tmp=0;
      while (filename[tmp]==' ')
        tmp++;
      db_strcpy(e->print_command,&filename[tmp]);
    }
    else if (db_strcmp(filename,"compilecommand:")==0)
    {
      fgets(filename,100,f);
      filename[db_strlen(filename)-1]='\0';
      tmp=0;
      while (filename[tmp]==' ')
        tmp++;
      db_strcpy(e->compile_command,&filename[tmp]);
    }
    else if (db_strcmp(filename,"makecommand:")==0)
    {
      fgets(filename,100,f);
      filename[db_strlen(filename)-1]='\0';
      tmp=0;
      while (filename[tmp]==' ')
        tmp++;
      db_strcpy(e->make_command,&filename[tmp]);
    }
    else if (db_strcmp(filename,"runcommand:")==0)
    {
      fgets(filename,100,f);
      filename[db_strlen(filename)-1]='\0';
      tmp=0;
      while (filename[tmp]==' ')
        tmp++;
      db_strcpy(e->run_command,&filename[tmp]);
    }
    else if (db_strcmp(filename,"compilererror:")==0)
    {
      fgets(filename,100,f);
      filename[db_strlen(filename)-1]='\0';
      tmp=0;
      while (filename[tmp]==' ')
        tmp++;
      e->compiler_error=(char**)db_realloc(e->compiler_error,(e->ce_count+1)*sizeof(char*));
      e->compiler_error[e->ce_count++]=db_strdup(&filename[tmp]);
    }
  }
  if (!e->ce_count)
  {
    e->compiler_error=(char**)db_realloc(e->compiler_error,(e->ce_count+1)*sizeof(char*));
    e->compiler_error[e->ce_count++]=db_strdup("%[^:]:%d");
    e->compiler_error=(char**)db_realloc(e->compiler_error,(e->ce_count+1)*sizeof(char*));
    e->compiler_error[e->ce_count++]=db_strdup("cc: \"%[^\"]\", line %d:");
  }
  fclose(f);
  ApplySettings(e);
EP;
} 
  
void SettingsOK(Widget w,Editor *e,void *cbs)
{
  char *tmpstr;
  XmStringTable items;
  int c,x;
  
SP("SettingsOK");
  XtVaGetValues(e->ai,XmNset,&e->auto_indent,NULL);
  XtVaGetValues(e->nw,XmNset,&e->new_window,NULL);
  XtVaGetValues(e->bu,XmNset,&e->blockundo,NULL);
  XtVaGetValues(e->ast,XmNset,&e->autosave,NULL);
  XtVaGetValues(e->ul,XmNvalue,&tmpstr,NULL);
  e->undolevels=atoi(tmpstr);
  if (e->undolevels<0)
    e->undolevels=0;
  XtFree(tmpstr);
  XtVaGetValues(e->asf,XmNvalue,&tmpstr,NULL);
  e->autosavetimeout=atoi(tmpstr);
  if (e->autosavetimeout<=0)
    e->autosavetimeout=1;
  XtFree(tmpstr);
  XtVaGetValues(e->pc,XmNvalue,&tmpstr,NULL);
  db_strncpy(e->print_command,tmpstr,99);
  XtFree(tmpstr);
  XtVaGetValues(e->cc,XmNvalue,&tmpstr,NULL);
  db_strncpy(e->compile_command,tmpstr,99);
  XtFree(tmpstr);
  XtVaGetValues(e->mc,XmNvalue,&tmpstr,NULL);
  db_strncpy(e->make_command,tmpstr,99);
  XtFree(tmpstr);
  XtVaGetValues(e->runc,XmNvalue,&tmpstr,NULL);
  db_strncpy(e->run_command,tmpstr,99);
  XtFree(tmpstr);
/*  XtVaGetValues(e->ce,XmNvalue,&tmpstr,NULL);
  db_strncpy(e->compiler_error,tmpstr,99);
  XtFree(tmpstr);*/
  if (e->ce_count)
  {
    for (x=0;x<e->ce_count;x++)
      db_free(e->compiler_error[x]);
    db_free(e->compiler_error);
    e->compiler_error=NULL;
    e->ce_count=0;
  }    
  XtVaGetValues(e->ce,XmNitemCount,&c,XmNitems,&items,NULL);
  if (c)
  {
    e->compiler_error=(char**)db_malloc(sizeof(char*)*c);
    e->ce_count=c;
    for (x=0;x<c;x++)
    {
      XmStringGetLtoR(items[x],XmSTRING_DEFAULT_CHARSET,&tmpstr);
      e->compiler_error[x]=db_strdup(tmpstr);
      XtFree(tmpstr);
    }
  }
  ApplySettings(e);
  SaveSettings(e);
  SetCurrentWidget(e->text);
EP;
}

void SettingsCancel(Widget w,Editor *e,void *cbs)
{
  SetCurrentWidget(e->text);
}

void AddCE(Widget w,Editor *e,void *cbs)
{
  char *str="";
  XmString xstr;
  int *pos,count;
  
  SP("AddCE");
#ifndef LESSTIFF    
  XtVaGetValues(XtNameToWidget(e->sd,"*CEField"),XmNvalue,&str,NULL);
#endif
  if (str[0])
  {
    xstr=XmStringCreateLocalized(str);
    if (XmListGetSelectedPos(e->ce,&pos,&count))
    {
      XmListAddItem(e->ce,xstr,pos[0]+1);
      XmListSelectPos(e->ce,pos[0]+1,True);
      XmListSetBottomPos(e->ce,pos[0]+1);
      XmStringFree(xstr);
      XtFree((char*)pos);
    }
    else
    {
      XmListAddItem(e->ce,xstr,0);
      XmListSelectPos(e->ce,0,True);
      XmListSetBottomPos(e->ce,0);
      XmStringFree(xstr);
    }    
  }
  XtFree(str);
  EP;
}

void ModifyCE(Widget w,Editor *e,void *cbs)
{
  char *str="";
  XmString xstr;
  int *pos,count;
  
  SP("ModifyCE");
#ifndef LESSTIFF    
  XtVaGetValues(XtNameToWidget(e->sd,"*CEField"),XmNvalue,&str,NULL);
#endif
  if (str[0])
  {
    if (XmListGetSelectedPos(e->ce,&pos,&count))
    {
      xstr=XmStringCreateLocalized(str);
      XmListReplaceItemsPos(e->ce,&xstr,1,pos[0]);
      XmListSelectPos(e->ce,pos[0],True);
      XmStringFree(xstr);
      XtFree((char*)pos);
    }
  }
  XtFree(str);
  EP;
}

void RemoveCE(Widget w,Editor *e,void *cbs)
{
  int *pos,count;
  
  SP("RemoveCE");
  if (XmListGetSelectedPos(e->ce,&pos,&count))
  {
    XmListDeletePos(e->ce,pos[0]);
    XtVaGetValues(e->ce,XmNitemCount,&count,NULL);
    if (pos[0]>count)
    {
      if (!count)
      {
#ifndef LESSTIFF    
        XtVaSetValues(XtNameToWidget(e->sd,"*CEField"),XmNvalue,"",NULL);
#endif
      }
      else
      {
        XmListSelectPos(e->ce,0,True);
        XmListSetBottomPos(e->ce,0);
      }
    }
    else
    {
      XmListSelectPos(e->ce,pos[0],True);
      XmListSetBottomPos(e->ce,pos[0]);
    }
    XtFree((char*)pos);
  }
  EP;
}

void SelectCE(Widget w,Editor *e,void *cbs)
{
  int count;
  XmStringTable items;
  char *str;
  
  SP("SelectCE");
  XtVaGetValues(e->ce,XmNselectedItemCount,&count,XmNselectedItems,&items,NULL);
  if (count)
  {
    XmStringGetLtoR(items[0],XmSTRING_DEFAULT_CHARSET,&str);
#ifndef LESSTIFF    
    XtVaSetValues(XtNameToWidget(e->sd,"*CEField"),XmNvalue,str,NULL);
#endif
    XtFree(str);
  }
  else
  {
#ifndef LESSTIFF    
    XtVaSetValues(XtNameToWidget(e->sd,"*CEField"),XmNvalue,"",NULL);
#endif
  }    
  EP;
}

void Settings(Editor *e)
{
  char tmpstr[100];
  int x;
  XmString xstr;
  Widget w,rc;
  Arg arg[2];
  
SP("Settings");
  if (e->sd==NULL)
  {
    e->sd=XmCreateMessageDialog(e->shell,"Settings",NULL,0);
    XtAddCallback(e->sd,XmNokCallback,(XtCallbackProc)SettingsOK,e);
    XtAddCallback(e->sd,XmNcancelCallback,(XtCallbackProc)SettingsCancel,e);
    XtAddCallback(e->sd,XmNhelpCallback,(XtCallbackProc)Help,"Settings");
    e->sf=XtVaCreateManagedWidget("settingsform",xmFormWidgetClass,e->sd,
    	XmNfractionBase,2,
    	NULL);
    e->ai=XtVaCreateManagedWidget("AutoIndent",xmToggleButtonWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_FORM,
    	XmNleftAttachment,XmATTACH_FORM,
	NULL);
    e->nw=XtVaCreateManagedWidget("NewWindow",xmToggleButtonWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->ai,
    	XmNleftAttachment,XmATTACH_FORM,
	NULL);
    e->bu=XtVaCreateManagedWidget("BlockUndo",xmToggleButtonWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->nw,
    	XmNleftAttachment,XmATTACH_FORM,
	NULL);
    e->ul=XtVaCreateManagedWidget("UndoLevels",xmTextFieldWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->bu,
    	XmNleftAttachment,XmATTACH_POSITION,
    	XmNleftPosition,1,
    	XmNrightAttachment,XmATTACH_FORM,
	NULL);
    e->ull=XtVaCreateManagedWidget("UndoLevelLabel",xmLabelGadgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->bu,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_POSITION,
    	XmNrightPosition,1,
    	XmNbottomAttachment,XmATTACH_OPPOSITE_WIDGET,
    	XmNbottomWidget,e->ul,
	NULL);
    e->pc=XtVaCreateManagedWidget("PrintCommand",xmTextFieldWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->ul,
    	XmNleftAttachment,XmATTACH_POSITION,
    	XmNleftPosition,1,
    	XmNrightAttachment,XmATTACH_FORM,
	NULL);
    e->pcl=XtVaCreateManagedWidget("PrintCommLabel",xmLabelGadgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->ull,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_POSITION,
    	XmNrightPosition,1,
    	XmNbottomAttachment,XmATTACH_OPPOSITE_WIDGET,
    	XmNbottomWidget,e->pc,
	NULL);
    e->cc=XtVaCreateManagedWidget("CompileCommand",xmTextFieldWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->pc,
    	XmNleftAttachment,XmATTACH_POSITION,
    	XmNleftPosition,1,
    	XmNrightAttachment,XmATTACH_FORM,
	NULL);
    e->ccl=XtVaCreateManagedWidget("CompileCommLabel",xmLabelGadgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->pcl,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_POSITION,
    	XmNrightPosition,1,
    	XmNbottomAttachment,XmATTACH_OPPOSITE_WIDGET,
    	XmNbottomWidget,e->cc,
	NULL);
    e->mc=XtVaCreateManagedWidget("MakeCommand",xmTextFieldWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->cc,
    	XmNleftAttachment,XmATTACH_POSITION,
    	XmNleftPosition,1,
    	XmNrightAttachment,XmATTACH_FORM,
	NULL);
    e->mcl=XtVaCreateManagedWidget("MakeCommLabel",xmLabelGadgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->ccl,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_POSITION,
    	XmNrightPosition,1,
    	XmNbottomAttachment,XmATTACH_OPPOSITE_WIDGET,
    	XmNbottomWidget,e->mc,
	NULL);
    e->runc=XtVaCreateManagedWidget("RunCommand",xmTextFieldWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->mc,
    	XmNleftAttachment,XmATTACH_POSITION,
    	XmNleftPosition,1,
    	XmNrightAttachment,XmATTACH_FORM,
	NULL);
    e->runcl=XtVaCreateManagedWidget("RunCommLabel",xmLabelGadgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->mcl,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_POSITION,
    	XmNrightPosition,1,
    	XmNbottomAttachment,XmATTACH_OPPOSITE_WIDGET,
    	XmNbottomWidget,e->runc,
	NULL);
    e->cel=XtVaCreateManagedWidget("CompileErrorLabel",xmLabelGadgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->runcl,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_NONE,
    	XmNbottomAttachment,XmATTACH_NONE,
	NULL);
    rc=XtVaCreateManagedWidget("CERC",xmRowColumnWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->cel,
    	XmNrightAttachment,XmATTACH_FORM,
    	XmNbottomAttachment,XmATTACH_NONE,
    	XmNorientation,XmVERTICAL,
	NULL);
    w=XtVaCreateManagedWidget("Add",xmPushButtonGadgetClass,rc,NULL);
    XtAddCallback(w,XmNactivateCallback,(XtCallbackProc)AddCE,(void*)e);
    w=XtVaCreateManagedWidget("Modify",xmPushButtonGadgetClass,rc,NULL);
    XtAddCallback(w,XmNactivateCallback,(XtCallbackProc)ModifyCE,(void*)e);
    w=XtVaCreateManagedWidget("Delete",xmPushButtonGadgetClass,rc,NULL);
    XtAddCallback(w,XmNactivateCallback,(XtCallbackProc)RemoveCE,(void*)e);
    XtSetArg(arg[0],XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE);
    e->ce=XmCreateScrolledList(e->sf,"CompileError",arg,1);
    XtManageChild(e->ce);
    XtVaSetValues(XtParent(e->ce),
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,e->cel,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_WIDGET,
    	XmNrightWidget,rc,
	NULL);
    XtAddCallback(e->ce,XmNbrowseSelectionCallback,(XtCallbackProc)SelectCE,(void*)e);
    w=XtVaCreateManagedWidget("CEField",xmTextFieldWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,XtParent(e->ce),
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_OPPOSITE_WIDGET,
    	XmNrightWidget,e->ce,
	NULL);
    e->asf=XtVaCreateManagedWidget("AutoSave",xmTextFieldWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,w,
    	XmNleftAttachment,XmATTACH_POSITION,
    	XmNleftPosition,1,
    	XmNrightAttachment,XmATTACH_FORM,
	NULL);
    e->ast=XtVaCreateManagedWidget("AutoSaveToggle",xmToggleButtonWidgetClass,e->sf,
    	XmNtopAttachment,XmATTACH_WIDGET,
    	XmNtopWidget,w,
    	XmNleftAttachment,XmATTACH_FORM,
    	XmNrightAttachment,XmATTACH_POSITION,
    	XmNrightPosition,1,
    	XmNbottomAttachment,XmATTACH_OPPOSITE_WIDGET,
    	XmNbottomWidget,e->asf,
	NULL);
    XtAddCallback(e->ast,XmNvalueChangedCallback,(XtCallbackProc)ToggleAutoSave,e);
  }
  XtVaSetValues(e->ai,XmNset,e->auto_indent,NULL);
  XtVaSetValues(e->nw,XmNset,e->new_window,NULL);
  XtVaSetValues(e->bu,XmNset,e->blockundo,NULL);
  sprintf(tmpstr,"%d",e->undolevels);
  XtVaSetValues(e->ul,XmNvalue,tmpstr,NULL);
  XtVaSetValues(e->pc,XmNvalue,e->print_command,NULL);
  XtVaSetValues(e->cc,XmNvalue,e->compile_command,NULL);
  XtVaSetValues(e->mc,XmNvalue,e->make_command,NULL);
  XtVaSetValues(e->runc,XmNvalue,e->run_command,NULL);
/*  XtVaSetValues(e->ce,XmNvalue,e->compiler_error,NULL);*/
  XmListDeleteAllItems(e->ce);
  for (x=0;x<e->ce_count;x++)
  {
    xstr=XmStringCreateLocalized(e->compiler_error[x]);
    XmListAddItem(e->ce,xstr,0);
    XmStringFree(xstr);
  }
#ifndef LESSTIFF    
  XtVaSetValues(XtNameToWidget(e->sd,"*CEField"),XmNvalue,"",NULL);
#endif  
  XtVaSetValues(e->ast,XmNset,e->autosave,NULL);
  sprintf(tmpstr,"%d",e->autosavetimeout);
  XtVaSetValues(e->asf,XmNvalue,tmpstr,XmNsensitive,e->autosave,NULL);
  XtManageChild(e->sd);
EP;
}

